unit GenericTableSet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Menus, StdCtrls, ExtCtrls, DB, DBTables, TableUtils, TableSet, DataWin,
  DBLookup;

type
  TGenericTableSetForm = class(TForm)
    MainMenu1: TMainMenu;
    File1: TMenuItem;
    ExportTableToExcel1: TMenuItem;
    FastExcel971: TMenuItem;
    SlowExcel951: TMenuItem;
    ExportTabletoAscii1: TMenuItem;
    N1: TMenuItem;
    PrinterSetup1: TMenuItem;
    N2: TMenuItem;
    Exit1: TMenuItem;
    Tools1: TMenuItem;
    AdminOnly1: TMenuItem;
    BackupExportAllTables1: TMenuItem;
    RestoreImportAllTables1: TMenuItem;
    N4: TMenuItem;
    RecreateTablesDangerous1: TMenuItem;
    DeleteAllTablesDangerous1: TMenuItem;
    ToolPanel: TPanel;
    Label1: TLabel;
    TableCombo: TComboBox;
    Panel1: TPanel;
    PrinterSetupDialog1: TPrinterSetupDialog;
    procedure Exit1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure TableComboChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure Panel1Resize(Sender: TObject);
    procedure FastExcel971Click(Sender: TObject);
    procedure SlowExcel951Click(Sender: TObject);
    procedure PrinterSetup1Click(Sender: TObject);
    procedure ExportTabletoAscii1Click(Sender: TObject);
    procedure BackupExportAllTables1Click(Sender: TObject);
    procedure DeleteAllTablesDangerous1Click(Sender: TObject);
    procedure RecreateTablesDangerous1Click(Sender: TObject);
    procedure RestoreImportAllTables1Click(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  protected
    TableSet   : TTableSetHelper;
    dw         : TDataWinForm;
    procedure  InitTableCombo;
    procedure  FillComboTable(Table: TTable);
    procedure  TableStatus(Sender: TObject; const Msg: string; LastOperation: boolean); virtual;
    procedure  AppendTablePopup;
  public
  end;

var
  GenericTableSetForm: TGenericTableSetForm;

implementation

{$R *.DFM}

procedure TGenericTableSetForm.Exit1Click(Sender: TObject);
begin
   Close;
end;

procedure TGenericTableSetForm.FormShow(Sender: TObject);
begin
   InitTableCombo;
end;

procedure TGenericTableSetForm.InitTableCombo;
begin
   TableCombo.Items.Clear;
   ForEachTableIn(TableSet.DataGroup, FillComboTable);
   TableCombo.ItemIndex := TableCombo.Items.IndexOf('Bugs');
   TableComboChange(nil);
end;

procedure TGenericTableSetForm.FillComboTable(Table: TTable);
begin
   TableCombo.Items.AddObject(Table.TableName, Table);
end;

procedure TGenericTableSetForm.TableComboChange(Sender: TObject);
var
   i : integer;
begin
   i := TableCombo.ItemIndex;
   if i = -1 then exit;
   dw.AssignTable(TableCombo.Items.Objects[i] as TTable);
   dw.Show;
end;

procedure TGenericTableSetForm.FormCreate(Sender: TObject);
begin
   dw := TDataWinForm.Create(Self);
   dw.BorderStyle := bsNone;
   dw.Parent := Panel1;
   AppendTablePopup;

   TableSet := TTableSetHelper.Create;
   TableSet.StatusProc := TableStatus;
end;

procedure TGenericTableSetForm.Panel1Resize(Sender: TObject);
var
   c : TComponent;
begin
   c := FindComponent('DataWinForm');
   if c=nil then exit;
   
   dw.Left := 0;
   dw.Top  := 0;
   dw.Width := Panel1.Width;
   dw.Height := Panel1.Height;
end;

procedure TGenericTableSetForm.FastExcel971Click(Sender: TObject);
begin
   ExportTableToExcel(dw.CurTable, False);
end;

procedure TGenericTableSetForm.SlowExcel951Click(Sender: TObject);
begin
   ExportTableToExcel(dw.CurTable, True);
end;

procedure TGenericTableSetForm.PrinterSetup1Click(Sender: TObject);
begin
   PrinterSetupDialog1.Execute;
end;

procedure TGenericTableSetForm.ExportTabletoAscii1Click(Sender: TObject);
begin
   dw.ExporttoAscii1Click(nil);
end;


procedure TGenericTableSetForm.BackupExportAllTables1Click(Sender: TObject);
begin
   TableSet.BackupTables;
end;

procedure TGenericTableSetForm.RestoreImportAllTables1Click(Sender: TObject);
begin
   TableSet.RestoreTables;
   dw.CurTable.Refresh;
   dw.Refresh_Filter_Combos;
end;

procedure TGenericTableSetForm.DeleteAllTablesDangerous1Click(Sender: TObject);
begin
   dw.DataSource1.Enabled := False;
   TableSet.DeleteTables;
   InitTableCombo;
end;

procedure TGenericTableSetForm.RecreateTablesDangerous1Click(Sender: TObject);
var
   OrigDG : TComponent;
   TmpDG  : TComponent;
begin
   with TableSet do begin
      TmpDG := TComponentClass(DataGroup.ClassType).Create(Self);
      OrigDG := DataGroup;
      CloseTables;   // For the original DataGroup (global DataModule1)
      DataGroup := TmpDG;

      CreateTablesWithIdx;
      TmpDG.Free;

      DataGroup := OrigDG;
      OpenTables;

      dw.CurTable.Refresh;
      dw.IndexComboChange(nil);
   end;
end;

procedure TGenericTableSetForm.FormDestroy(Sender: TObject);
begin
   TableSet.Free;
end;

procedure TGenericTableSetForm.TableStatus(Sender: TObject; const Msg: string; LastOperation: boolean);
begin
   if not LastOperation then begin
      dw.DataSource1.Enabled := False;
   end else begin
      dw.DataSource1.Enabled := True;
   end;
end;

procedure CopySubMenus(SrcMenu: TMenuItem; DestMenu: TMenuItem);
var
   i  : integer;
   mi : TMenuItem;
begin
   with SrcMenu do begin
      for i := 0 to Count-1 do begin
         mi := TMenuItem.Create(DestMenu);
         mi.Caption := Items[i].Caption;
         mi.Hint    := Items[i].Hint;
         mi.OnClick := Items[i].OnClick;
         DestMenu.Add(mi);
         CopySubMenus(Items[i], mi);
      end;
   end;
end;

procedure TGenericTableSetForm.AppendTablePopup;
begin
   {Table1.OnClick := DataWinForm.TablePopupMenu.OnPopup;
   CopySubMenus(DataWinForm.TablePopupMenu.Items, Table1);}
end;


procedure TGenericTableSetForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
   dw.SaveColumnSettings;
end;

end.
