unit PoolMain;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, DBTables, ShellAPI, ShlObj, ExtCtrls;

type
  TMainForm = class(TForm)
    CreateBut: TButton;
    ViewBut: TButton;
    ExitBut: TButton;
    Panel1: TPanel;
    DeleteBut: TButton;
    procedure ExitButClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure CreateButClick(Sender: TObject);
    procedure DeleteButClick(Sender: TObject);
    procedure ViewButClick(Sender: TObject);
    procedure InitButClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure  CreateAlias;
    procedure  UpdateStatus;
    procedure  AppendText(Tab: TTable; const FileName: string);
    procedure  FillWeeksTable;
    procedure  AddAdminUser;
    procedure  UpdateGlobals;
  end;

var
  MainForm: TMainForm;

implementation

uses Pool_Mod_Unit, TabSet;

{$R *.DFM}

procedure TMainForm.ExitButClick(Sender: TObject);
begin
   Close;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
   Show;
   Update;
   
   Screen.Cursor := crHourGlass;
   try
      UpdateStatus;
   finally
      Screen.Cursor := crDefault;
   end;
end;

procedure TMainForm.CreateAlias;
begin
end;

procedure TMainForm.UpdateStatus;
var
   b : boolean;
   s : string;
   p : TStringList;
begin
   b := Session.IsAlias(POOL_ALIAS);

   CreateBut.Enabled := not b;
   ViewBut.Enabled   := b;
   DeleteBut.Enabled := b;

   if b then begin
      p := TStringList.Create;
      Session.GetAliasParams(POOL_ALIAS, p);
      s := Format('Alias = %s   Path = %s', [POOL_ALIAS, p.Values['PATH']]);
      p.Free;
   end else begin
      s := 'Pool Alias not defined';
   end;
   Panel1.Caption := s;
end;

procedure TMainForm.CreateButClick(Sender: TObject);
var
   lpbi       : TBrowseInfo;
   ItemIDList : PItemIDList;
   buf        : array[0..MAX_PATH] of char;
   rc         : boolean;
begin
   FillChar(lpbi, sizeof(lpbi), 0);
   lpbi.lpszTitle := 'Select a folder for the WABD Pool Database';
   ItemIDList := SHBrowseForFolder(lpbi);
   if ItemIDList = nil then exit;      // User hit Cancel
   rc := SHGetPathFromIDList(ItemIDList, buf);
   Assert(rc, 'SHGetPathFromIDList Failed!');

   Session.AddStandardAlias(POOL_ALIAS, buf, 'PARADOX');
   Session.SaveConfigFile;

   Pool_Module.ForEachTable(Pool_Module.SetDatabase);
   Pool_Module.CreateTables;
   UpdateStatus;

   InitButClick(nil);
end;


procedure TMainForm.DeleteButClick(Sender: TObject);
var
   mr : integer;
begin
   mr := MessageDlg('Are you SURE you want to delete the database', mtWarning,
      mbYesNoCancel, 0);
   if mr<>mrYes then exit;

   Panel1.Caption := 'Deleting Tables...';
   Panel1.Update;

   Pool_Module.ForEachTable(Pool_Module.SetDatabase);
   Pool_Module.DeleteTables;
   Session.DeleteAlias(POOL_ALIAS);
   Session.SaveConfigFile;

   UpdateStatus;
end;

procedure TMainForm.ViewButClick(Sender: TObject);
begin
   Pool_Module.InitSession;
   Pool_Module.ForEachTable(Pool_Module.SetDatabase);
   TableSetForm.Show;
end;

procedure TMainForm.InitButClick(Sender: TObject);
var
   TeamTab  : string;
   GameTab  : string;
   sl       : TStringList;
begin
   // Initialize the Session
   Pool_Module.Session1.SessionName := 'MainSession';
   sl := TStringList.Create;
   Pool_Module.Session1.GetAliasParams(POOL_ALIAS, sl);
   Pool_Module.Session1.NetFileDir  := sl.Values['PATH'];
   Pool_Module.Session1.Active := True;
   sl.Free;

   Pool_Module.ForEachTable(Pool_Module.SetDatabase);
   
   TeamTab := ExtractFilePath(Application.ExeName) + 'TeamTable.txt';
   GameTab := ExtractFilePath(Application.ExeName) + 'GameTable.txt';

   AppendText(Pool_Module.TeamsTable, TeamTab);
   AppendText(Pool_Module.GamesTable, GameTab);

   FillWeeksTable;
   AddAdminUser;
   UpdateGlobals;

   Pool_Module.LogAction(ADMIN_USER_ID, Log_CreateTables);
end;

procedure TMainForm.AppendText(Tab: TTable; const FileName: string);
var
   tf       : TStringList;
   flds     : TStringList;
   line     : TStringList;
   row      : integer;
   i        : integer;
   NumErr   : integer;
begin
   Tab.Active := True;

   tf := TStringList.Create;
   tf.LoadFromFile(FileName);

   flds := TStringList.Create;
   flds.CommaText := tf.Strings[0];

   line := TStringList.Create;

   NumErr := 0;
   for row := 1 to tf.Count-1 do begin
      Panel1.Caption := 'Appending Line ' + IntToStr(row) + ' from ' + FileName;
      Panel1.Update;

      Tab.Append;
      line.CommaText := tf.Strings[row];
      for i := 0 to line.Count-1 do begin
         Tab.FieldByName(flds.Strings[i]).AsString := line.Strings[i];
      end;
      try
         Tab.Post;
      except
         on e:Exception do NumErr := NumErr + 1;
      end;
   end;

   line.Free;
   flds.Free;
   tf.Free;

   Panel1.Caption := 'Number of Append Errors = ' + IntToStr(NumErr);
end;

// Move it back to a workday, and NLT close of business (4:30 pm)
procedure MoveBack(var dt: TDateTime);
const
   COB = 16.5 / 24.0;    // # of hours since midnight
var
   Good  : boolean;
begin
   repeat
      Good := (DayOfWeek(dt)<>1) and (DayOfWeek(dt)<>7);
      if not Good then begin
          dt := Trunc(dt) - (1 / (24.0 * 3600));   // 11:59 pm prev day
      end;
      if Frac(dt) > COB then begin
         dt := Trunc(dt) + COB;
      end;
   until Good;
end;

procedure TMainForm.FillWeeksTable;
var
   wk, oldwk   : integer;
   MinDate, dt : TDateTime;
   NumGame     : integer;
   Done        : boolean;
begin
   with Pool_Module.GamesTable do begin
      Active := True;
      First;

      oldwk   := -1;
      MinDate := 1e20;
      NumGame := 0;
      // while not EOF do begin
      Done := False;
      repeat
         wk := FieldByName('WeekID').AsInteger;
         if EOF then begin
            Done := True;
            wk   := -2;
         end;

         if (wk<>oldwk) and (oldwk<>-1) then begin
            // Process the OldWeek
            with Pool_Module.WeeksTable do begin
               Active := True;
               Append;
               FieldByName('WeekID').AsInteger := oldwk;
               FieldByName('TotGames').AsInteger := NumGame;
               MoveBack(MinDate);
               FieldByName('LockOutTime').AsDateTime := MinDate;
               try
                  Post;
               except
                  on e:Exception do begin end;
               end;
            end;

            // Reset Min's
            MinDate := 1e20;
            NumGame := 0;
         end;

         dt := FieldByName('GameTime').AsDateTime;
         if dt < MinDate then MinDate := dt;
         NumGame := NumGame + 1;

         oldwk := wk;

         Next;
      until Done;
   end;
end;

procedure TMainForm.AddAdminUser;
begin
   with Pool_Module.UserTable do begin
      Active := True;
      Append;
      FieldByName('UserID').AsInteger := ADMIN_USER_ID;
      FieldByName('UserName').AsString := 'Admin';
      FieldByName('UserPass').AsString := 'Admin';
      FieldByName('LastLogon').AsDateTime := Now;
      FieldByName('NumLogons').AsInteger := 0;
      Post;
   end;
end;

procedure TMainForm.UpdateGlobals;
begin
   with Pool_Module.GlobalsTable do begin
      Active := True;
      Append;
      FieldByName('CurWeek').AsInteger := 1;
      FieldByName('EntryCost').AsFloat := 2.0;
      Post;
   end;
end;


end.
