unit WABD_Objects;

interface

uses Classes, SysUtils, Forms, Graphics, Controls, StdCtrls, ExtCtrls, DB,
      Grids, Windows, ISAPI, WABD_About, SafeBitmap, Compat;

{$DEFINE NOT_WABD_REGISTERED}      // Change this to $DEFINE for registered users

{$IFDEF VER100}
{$ASSERTIONS ON}
{$ENDIF}

type
   // ************************************************************************
   // "Base" Level objects

   TWABD_Session = class;
   TWABD_Form = class;
   TWABD_Parent = class;
   TWABD_SessionMgr = class;
   TWABD_Base_Image = class;
   TWABD_Image = class;
   TWABD_SectionObject = class;
   TWABD_Button = class;
   TWABD_SectionObjectClass = class of TWABD_SectionObject;

   TGarbageThread = class(TThread)
   protected
      procedure   DoGarbageCollection;
      procedure   Execute; override;
   public
      SesMgr      : TWABD_SessionMgr;
   end;

   TWABD_OnCreateSession = procedure(var NewSession: TWABD_Session) of object;
   TWABD_OnDestroySession = procedure(Session: TWABD_Session) of object;

   TWABD_SessionMgr = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FCreate     : TWABD_OnCreateSession;
      FDestroy    : TWABD_OnDestroySession;
      FCheck      : integer;        // Interval to perform Garbage Collection (seconds)
      FGarbage    : boolean;        // True = perform Garbage collection
      GarbageThrd : TGarbageThread;
      StopEvent   : THandle;
      NewSesID    : integer;        // Increment by 1 for each new session
      DLLName     : string;
      LogStrings  : TStringList;
      LogCS       : TRTLCriticalSection;
      SesMgrCS    : TRTLCriticalSection;
      CreateTime  : TDateTime;
      function    CreateNewSession: string;
      function    RunExistingSession(const IdStr: string; InputList: TStrings): string;
      procedure   CheckLogOff(Ses: TWABD_Session);
      procedure   LogUserInfo(var ECB: TEXTENSION_CONTROL_BLOCK);
      procedure   AddToLog(const s: string);
      function    FormatStatus: string;
      procedure   DoDestroySession(Ses: TWABD_Session);
   public
      SessionList : TThreadList;
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    OnFormSubmit(var ECB: TEXTENSION_CONTROL_BLOCK): string;
   published
      property    CheckTimeoutInterval: integer read FCheck write FCheck;
      property    GarbageCollection: boolean read FGarbage write FGarbage;
      property    OnCreateSession: TWABD_OnCreateSession read FCreate write FCreate;
      property    OnDestroySession: TWABD_OnDestroySession read FDestroy write FDestroy;
      property    About: TWABDAbout read FAbout write FAbout;
   end;


   TWABD_OnException = procedure(E: Exception; var Handled: boolean) of object;

   TWABD_Session = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FMainForm   : TWABD_Form;
      FLogon      : TNotifyEvent;
      FLogOff     : TNotifyEvent;
      FTimeOut    : TNotifyEvent;
      DidLogOff   : boolean;
      FTimeLen    : integer;
      FSessionID  : integer;
      CreateTime  : TDateTime;
      LastAccess  : TDateTime;
      FCurForm    : TWABD_Form;
      NewForm     : TWABD_Form;     // To "Show" new forms
      FExcept     : TWABD_OnException;
      FDLLName    : string;
      LocalError  : string;         // set by "ProcessInput" if the current form can't set it's properties or call it's handler 
      function    ProcessInput(InputList: TStrings): string;
      // procedure   Lock;
      // procedure   Unlock;
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   Loaded; override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      property    SessionID: integer read FSessionID;
      property    CurForm: TWABD_Form read FCurForm;
      procedure   LogOff;
   published
      property    MainForm: TWABD_Form read FMainForm write FMainForm;
      property    TimeOutLength: integer read FTimeLen write FTimeLen;
      property    OnFirstLogon: TNotifyEvent read FLogon write FLogon;
      property    OnLogOff: TNotifyEvent read FLogOff write FLogOff;
      property    OnTimeOut: TNotifyEvent read FTimeOut write FTimeOut;
      property    OnException: TWABD_OnException read FExcept write FExcept;
      property    DLLName: string read FDLLName;
      property    About: TWABDAbout read FAbout write FAbout;
   end;

   TWABD_ObjectClass = class of TWABD_Object;
   TWABD_Object = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FParent     : TWABD_Parent;
      FParentName : string;      // tmp to hold name before all components are loaded
      FOnChange   : TNotifyEvent;
      FVisible    : boolean;
      FOrder      : integer;     // Only used on loading
      InLoaded    : boolean;
      procedure   SetParent(NewParent: TWABD_Parent);
      procedure   SetVisible(b: boolean);
      procedure   Changed;
      procedure   DefineProperties(Filer: TFiler); override;
      procedure   Loaded; override;
      procedure   ReadParentName(Reader: TReader);
      procedure   WriteParentName(Writer: TWriter);
      function    GetOrder: integer;
      procedure   SetOrder(NewOrder: integer);
      procedure   SetName(const Value: TComponentName); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; virtual; abstract;
      function    Object_To_Control(AOwner: TWinControl): TControl; virtual; abstract;
      procedure   HTML_To_Object(FormVal: string); virtual; abstract;
      function    GetParentForm: TWABD_Form;
      property    OnChange: TNotifyEvent read FOnChange write FOnChange;
      property    Parent: TWABD_Parent read FParent write SetParent;
      property    Order: integer read GetOrder write SetOrder;
   published
      property    Visible: boolean read FVisible write SetVisible;
      property    About: TWABDAbout read FAbout write FAbout;
   end;

   TWABD_ForEach = procedure(Child: TWABD_Object; var Stop: boolean; UserData: pointer) of object;

   TWABD_Parent = class(TWABD_Object)
   protected
      FWABD_Objs  : TList;
      tmp         : string;
      TheChild    : TWABD_Object;
      function    GetWABDObjects(i: integer): TWABD_Object;
      function    GetWABDObjCount: integer;
      function    ButtonByCaption(Caption: string): TWABD_Button;
      function    GetDefaultButton: TWABD_Button;
      procedure   DefButProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
      procedure   ChildNameProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
      procedure   ButCapProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
      procedure   ChildChanged(Sender: TObject); virtual;
      procedure   Loaded; override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      function    ForEachChild(ForEachProc: TWABD_ForEach; UserData: pointer): boolean;
      property    Children[i: integer] : TWABD_Object read GetWABDObjects;
      property    ChildCount: integer read GetWABDObjCount;
      function    ChildByName(ChildName: string): TWABD_Object;
   end;


   // ************************************************************************
   // "Top" Level objects

   TWABD_ReflectNotify = procedure(Sender: TObject; AComponent: TComponent; Operation: TOperation) of object;
   TChildNameChangedProc = procedure(Sender: TObject; const OldName, NewName: string) of object;

   TWABDEditForm = class(TPersistent)
   public
      ParForm     : TWABD_Form;
   end;

   TWABD_Form = class(TWABD_Parent)
   protected
      FEdForm     : TWABDEditForm;
      FOnShow     : TNotifyEvent;
      FOnCreate   : TNotifyEvent;
      FWSession   : TWABD_Session;
      FSesID      : integer;
      FTextColor  : TColor;
      FLinkColor  : TColor;
      FVLinkColor : TColor;
      FBgndColor  : TColor;
      FBgrdImage  : TWABD_Image;
      FTitle      : string;
      FNameChge   : TChildNameChangedProc;
      function    GetFormBody: string;
      function    GetNagStr: string;
      function    GetLocalError: string;
      function    DoMain: string;
      procedure   UncheckRadioButs(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
      procedure   ParseImageParams(Inputs: TStrings; var ImageName: string; var x,y: integer);
      procedure   SetProperties(InputList: TStrings);
      procedure   Call_Handler(InputList: TStrings);
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetEdForm(NewEdForm: TWABDEditForm);
   public
      RefNotify   : TWABD_ReflectNotify;                    // No one should use this but it's Component Editor
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      procedure   ChildChanged(Sender: TObject); override;
      function    Object_To_HTML: string; override;
      procedure   HTML_To_Object(FormVal: string); override;
      procedure   Show;
      property    OnChildNameChanged: TChildNameChangedProc read FNameChge write FNameChge;
   published
      property    TextColor: TColor read FTextColor write FTextColor default clNone;
      property    LinkColor: TColor read FLinkColor write FLinkColor default clNone;
      property    VLinkColor: TColor read FVLinkColor write FVLinkColor default clNone;
      property    BgndColor: TColor read FBgndColor write FBgndColor default clNone;
      property    OnShow: TNotifyEvent read FOnShow write FOnShow;
      property    OnCreate: TNotifyEvent read FOnCreate write FOnCreate;
      property    BgrdImage: TWABD_Image read FBgrdImage write FBgrdImage;
      property    FormTitle: string read FTitle write FTitle;
      property    Session: TWABD_Session read FWSession write FWSession;
      property    EditForm: TWABDEditForm read FEdForm write SetEdForm;
   end;


   TWABD_FormSection_Base = class(TWABD_Parent)
   private
      NumRow      : integer;
      NumCol      : integer;
      ColSizes    : array[0..255] of integer;   // eg 5,5,6,6
      RowSizes    : array[0..255] of integer;
      ColTot      : array[0..255] of integer;   // eg 5,10,16,22
      RowTot      : array[0..255] of integer;
      FindCol     : integer;                    // Used by ControlAtFunc
      FindRow     : integer;
      FindCon     : TWABD_SectionObject;
   protected
      FGridX      : integer;
      FGridY      : integer;
      FCellBorder : integer;
      FCellSpace  : integer;
      FCellPad    : integer;
      function    DoFormSection: string;
      procedure   SetGridX(NewX: integer);
      procedure   SetGridY(NewY: integer);
      procedure   AutoSizeRowCol;
      procedure   ControlAtFunc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
      function    AddControl(ControlClass: TWABD_SectionObjectClass; Col, Row: integer): TWABD_SectionObject;
      function    ControlAtPos(Col, Row: integer): TWABD_SectionObject;
      property    GridX: integer read FGridX write SetGridX;
      property    GridY: integer read FGridY write SetGridY;
      property    CellBorder: integer read FCellBorder write FCellBorder;
      property    CellSpacing: integer read FCellSpace write FCellSpace;
      property    CellPadding: integer read FCellPad write FCellPad;
   end;

   TWABD_FormSection = class(TWABD_FormSection_Base)
   published
      property    GridX;
      property    GridY;
      property    CellBorder;
   end;


   TWABD_Header = class(TWABD_Object)
   protected
      FNum        : integer;
      FCaption    : string;
      procedure   SetNum(NewNum: integer);
      procedure   SetName(const Value: TComponentName); override;
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    HeaderNum: integer read FNum write SetNum;
      property    Caption: string read FCaption write FCaption;
   end;


   TWABD_Raw = class(TWABD_Object)
   protected
      FLines      : TStrings;
      procedure   SetLines(NewLines: TStrings);
      procedure   SetName(const Value: TComponentName); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Lines: TStrings read FLines write SetLines;
   end;


   TWABDTableClick = procedure(Sender: TObject; RowIndex: integer) of object;

   TWABD_BaseTable = class(TWABD_Object)
   protected
      FCellBorder : integer;
      FCellSpace  : integer;
      FCanClick   : boolean;
      FClickText  : string;
      FOnClick    : TWABDTableClick;
   published
      constructor Create(AOwner: TComponent); override;
      property    CellBorder: integer read FCellBorder write FCellBorder;
      property    CellSpacing: integer read FCellSpace write FCellSpace;
      property    CanClick: boolean read FCanClick write FCanClick;
      property    ClickText: string read FClickText write FClickText;
      property    OnClick: TWABDTableClick read FOnClick write FOnClick;
   end;

   TStringArray = array[0..0] of string;
   PStringArray = ^TStringArray;

   TWABD_Table_Strings = class(TPersistent)
   protected
      XSize       : integer;
      YSize       : integer;
      FData       : PStringArray;
      function    GetString(x,y: integer): string;
      procedure   SetString(x,y: integer; NewString: string);
      procedure   DefineProperties(Filer: TFiler); override;
      procedure   WriteProps(Writer: TWriter);
      procedure   ReadProps(Reader: TReader);
      procedure   FreeData;
      procedure   SetSize(x,y: integer);
   public
      constructor Create;
      destructor  Destroy; override;
      procedure   Assign(Source: TPersistent); override;
      property    Strings[x,y: integer]: string read GetString write SetString; default;
      procedure   SafeSetSize(x,y: integer);
      property    Cols: integer read XSize;
      property    Rows: integer read YSize;
   end;

   TStringGridEx = class(TStringGrid)
   public
      ClickText   : string;
      CanClick    : boolean;
   end;

   TWABD_Table = class(TWABD_BaseTable)
   protected
      // CellData
      FFixCol     : integer;
      FFixRow     : integer;
      FStrings    : TWABD_Table_Strings;
      procedure   SetStrings(NewStrings: TWABD_Table_Strings);
      function    GetBut(Row: integer): string;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    FixedCols: integer read FFixCol write FFixCol;
      property    FixedRows: integer read FFixRow write FFixRow;
      property    Cells: TWABD_Table_Strings read FStrings write SetStrings;
   end;


   TWABD_DataTable = class(TWABD_Parent)
   protected
      FDataLink   : TDataLink;
      FFormSec    : TWABD_FormSection;
      FNavForm    : TWABD_FormSection;
      FTable      : TWABD_Table;
      FShowForm   : boolean;
      FShowTable  : boolean;
      FNavButs    : boolean;
      FMaxRows    : integer;
      FColWidth   : integer;
      FNumCol     : integer;
      Stat        : string;
      DidAppend   : boolean;
      procedure   InitForm;
      procedure   InitTable;
      procedure   InitNavButs;
      function    FieldTypeIsString(ft: TFieldType): boolean;
      procedure   CreateNavBut(x, y : integer; ButCap: string; OnClick: TNotifyEvent);
      procedure   FirstClick(Sender: TObject);
      procedure   LastClick(Sender: TObject);
      procedure   NextClick(Sender: TObject);
      procedure   PrevClick(Sender: TObject);
      procedure   NextPgClick(Sender: TObject);
      procedure   PrevPgClick(Sender: TObject);
      procedure   AddClick(Sender: TObject);
      procedure   EditClick(Sender: TObject);
      procedure   DeleteClick(Sender: TObject);
      function    GetDataSource: TDataSource;
      procedure   SetDataSource(NewDataSource: TDataSource);
      procedure   Notification(AComponent: TComponent; Operation: TOperation); override;
      procedure   SetName(const NewName: TComponentName); override;
      procedure   TableClick(Sender: TObject; RowIndex: integer);
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      procedure   HTML_To_Object(FormVal: string); override;
      property    FormSection: TWABD_FormSection read FFormSec;
      property    NavSection: TWABD_FormSection read FNavForm;
      property    Table: TWABD_Table read FTable;
   published
      property    DataSource: TDataSource read GetDataSource write SetDataSource;
      property    ShowEditForm: boolean read FShowForm write FShowForm;
      property    ShowTable: boolean read FShowTable write FShowTable;
      property    ShowNavButs: boolean read FNavButs write FNavButs;
      property    MaxRows: integer read FMaxRows write FMaxRows;
      property    FormColWidth: integer read FColWidth write FColWidth;
      property    NumCols: integer read FNumCol write FNumCol;
   end;


   TWABD_Hidden = class(TWABD_Object)
   protected
      FValue      : string;
   public
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Value: string read FValue write FValue;
   end;


   TWABD_BlankLines = class(TWABD_Object)
   protected
      FNumLines   : integer;
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    NumLines: integer read FNumLines write FNumLines;
   end;


   // ************************************************************************
   // "Form" Level objects (Buttons, Labels, Edit boxes, etc)

   TWABD_SectionObject = class(TWABD_Object)
   private
      Row, Col    : integer;  // Used by FormSection_Base's AutoSizeRowCol function
      OrigLeft    : integer;
      OrigTop     : integer;
   protected
      FLeftPos    : integer;  // Table start Pos (Left, Top)
      FTopPos     : integer;
      FWidth      : integer;  // Table Span Col & Row (Width, Height)
      FHeight     : integer;
      FColSpan    : integer;  // if -1, uses Width/GridX for ColSpan, else this number
      FRowSpan    : integer;
      procedure   SetLeft(NewLeft: integer);
      procedure   SetTop(NewTop: integer);
   public
      property    Width: integer read FWidth;
      property    Height: integer read FHeight;
   published
      constructor Create(AOwner: TComponent); override;
      property    LeftPos: integer read FLeftPos write SetLeft;
      property    TopPos: integer read FTopPos write SetTop;
      property    ColSpan: integer read FColSpan write FColSpan default -1;
      property    RowSpan: integer read FRowSpan write FRowSpan default -1;
   end;

   TWABD_LinesObject = class(TWABD_SectionObject)
   protected
      FLines      : TStrings;
      procedure   SetLines(NewLines: TStrings);
      procedure   SetName(const Value: TComponentName); override;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
   published
      property    Lines: TStrings read FLines write SetLines;
   end;


   TWABD_Anchor = class(TWABD_SectionObject)
   protected
      FDest       : string;
      FCaption    : string;
      FBold       : boolean;
      FItalic     : boolean;
      FUnderline  : boolean;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetCaption(NewCaption: string);
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Caption: string read FCaption write SetCaption;
      property    Destination: string read FDest write FDest;
      property    Bold: boolean read FBold write FBold;
      property    Italic: boolean read FItalic write FItalic;
      property    Underline: boolean read FUnderline write FUnderline;
   end;


   TWABD_MouseDown = procedure(Sender: TObject; X, Y: integer) of object;

   TWABD_HotSpot = class(TComponent)
   protected
      FAbout      : TWABDAbout;
      FX1         : integer;
      FY1         : integer;
      FX2         : integer;
      FY2         : integer;
      FImParent   : TWABD_Base_Image;
      FOnClick    : TNotifyEvent;
      FChange     : TNotifyEvent;
      procedure   SetX1(i: integer);
      procedure   SetY1(i: integer);
      procedure   SetX2(i: integer);
      procedure   SetY2(i: integer);
      procedure   SetName(const Value: TComponentName); override;
      procedure   Changed; virtual;
   public
      destructor  Destroy; override;
      property    OnChange: TNotifyEvent read FChange write FChange;
   published
      property    X1: integer read FX1 write SetX1;
      property    Y1: integer read FY1 write SetY1;
      property    X2: integer read FX2 write SetX2;
      property    Y2: integer read FY2 write SetY2;
      property    ImageParent: TWABD_Base_Image read FImParent write FImParent;
      property    OnClick: TNotifyEvent read FOnClick write FOnClick;
      property    About: TWABDAbout read FAbout write FAbout;
   end;

   TWABD_HotSpots = class(TPersistent)
   public
      ParImage    : TWABD_Base_Image;
   end;

   TWABD_Base_Image = class(TWABD_SectionObject)
   protected
      FAltText    : string;
      FMouseDown  : TWABD_MouseDown;
      FOnClick    : TNotifyEvent;
      FImagePath  : string;
      FLocalPath  : string;
      FHotSpots   : TWABD_HotSpots;
      FImgWidth   : integer;
      FImgHeight  : integer;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetImagePath(NewImagePath: string);
      procedure   SetLocalPath(NewLocalPath: string);
      procedure   Loaded; override;
      procedure   UpdateImageSize;
      procedure   SetImgWidth(w: integer); virtual;
      procedure   SetImgHeight(h: integer); virtual;
      procedure   MouseDown(x, y: integer); virtual;
      procedure   SetHotSpots(HS: TWABD_HotSpots);
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
      property    ImagePath: string read FImagePath write SetImagePath;
   published
      property    AltText: string read FAltText write FAltText;
      property    OnMouseDown: TWABD_MouseDown read FMouseDown write FMouseDown;
      property    OnClick: TNotifyEvent read FOnClick write FOnClick;
      property    LocalPath: string read FLocalPath write SetLocalPath;
      property    ImageWidth: integer read FImgWidth write SetImgWidth;
      property    ImageHeight: integer read FImgHeight write SetImgHeight;
      property    HotSpots: TWABD_HotSpots read FHotSpots write SetHotSpots;
   end;

   TWABD_Image = class(TWABD_Base_Image)
   published
      property    ImagePath;
   end;

   TWABD_LiveImage = class(TWABD_Base_Image)
   protected
      FSafeBmp    : TSafeBitmap;
      FDirty      : boolean;
      FCanvas     : TCanvas;
      FFileName   : string;
      WroteFile   : boolean;
      procedure   SetImgWidth(w: integer); override;
      procedure   SetImgHeight(h: integer); override;
      function    GetSafeBitmap: TSafeBitmap;
      function    GetCanvas: TCanvas;
      procedure   Loaded; override;
      function    GetFileName: string;
      function    GetNewName: string;
      procedure   SaveImage;
   public
      constructor Create(AOwner: TComponent); override;
      destructor  Destroy; override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   ReleaseCanvas;
      property    SafeBitmap: TSafeBitmap read GetSafeBitmap;
      property    Dirty: boolean read FDirty;
      property    Canvas: TCanvas read GetCanvas;
      property    FileName: string read GetFileName;
   end;


   TWABD_Label = class(TWABD_SectionObject)
   protected
      FCaption    : string;
      FBold       : boolean;
      FItalic     : boolean;
      FUnderline  : boolean;
      FFontColor  : TColor;
      FFontSize   : integer;
      FCanClick   : boolean;
      FOnClick    : TNotifyEvent;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetCaption(NewCaption: string);
      procedure   SetFontSize(NewSize: integer);
      procedure   UpdateWidHgt;
      procedure   SetBold(NewBold: boolean);
      function    GetHRef: string;
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Caption: string read FCaption write SetCaption;
      property    FontColor: TColor read FFontColor write FFontColor default clNone;
      property    Bold: boolean read FBold write SetBold default False;
      property    Italic: boolean read FItalic write FItalic default False;
      property    Underline: boolean read FUnderline write FUnderline default False;
      property    FontSize: integer read FFontSize write SetFontSize default 3;
      property    CanClick: boolean read FCanClick write FCanClick;
      property    OnClick: TNotifyEvent read FOnClick write FOnClick;
   end;


   TWABD_Memo = class(TWABD_LinesObject)
   protected
      FCols       : integer;
      FRows       : integer;
      procedure   SetCols(NewCols: integer);
      procedure   SetRows(NewRows: integer);
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Cols: integer read FCols write SetCols;
      property    Rows: integer read FRows write SetRows;
   end;


   TWABD_Button = class(TWABD_SectionObject)
   protected
      FCaption    : string;
      FOnClick    : TNotifyEvent;
      FDefault    : boolean;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetCaption(NewCaption: string);
   public
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Caption: string read FCaption write SetCaption;
      property    OnClick: TNotifyEvent read FOnClick write FOnClick;
      property    Default: boolean read FDefault write FDefault default False;
   end;


   TWABD_Edit = class(TWABD_SectionObject)
   protected
      FText       : string;
      FPass       : boolean;
      FSize       : integer;
      FMax        : integer;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetSize(NewSize: integer);
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Text: string read FText write FText;
      property    IsPassword: boolean read FPass write FPass;
      property    Size: integer read FSize write SetSize;
      property    MaxLength: integer read FMax write FMax;
   end;


   TWABD_ComboBox = class(TWABD_LinesObject)
   protected
      FSelIndex   : integer;
   public
      constructor Create(AOwner: TComponent); override;
      property    SelIndex : Integer read FSelIndex write FSelIndex;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   end;


   TWABD_RadioButton = class(TWABD_SectionObject)
   protected
      FCaption    : string;
      FCheck      : boolean;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetCaption(NewCaption: string);
   public
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Caption: string read FCaption write SetCaption;
      property    Checked: boolean read FCheck write FCheck;
   end;


   TWABD_ListBox = class(TWABD_LinesObject)
   protected
      FSize       : integer;
      FSelIndex   : integer;
      procedure   SetSize(NewSize: integer);
   public
      constructor Create(AOwner: TComponent); override;
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
      property    SelIndex: integer read FSelIndex write FSelIndex;
   published
      property    Size: integer read FSize write SetSize;
   end;


   TWABD_CheckBox = class(TWABD_SectionObject)
   protected
      FCaption    : string;
      FCheck      : boolean;
      procedure   SetName(const Value: TComponentName); override;
      procedure   SetCaption(NewCaption: string);
   public
      function    Object_To_HTML: string; override;
      function    Object_To_Control(AOwner: TWinControl): TControl; override;
      procedure   HTML_To_Object(FormVal: string); override;
   published
      property    Caption: string read FCaption write SetCaption;
      property    Checked: boolean read FCheck write FCheck;
   end;


   // ************************************************************************
   // Misc. Helper objects

   TPaintPanel = class(TPanel)
   protected
      FFormSec    : TWABD_FormSection_Base;
      FDrawPic    : boolean;
      FCellBord   : integer;
      FDMode      : boolean;
      procedure   Paint; override;
      procedure   SetDrawPic(b: boolean);
      procedure   SetCellBorder(nb: integer);
      procedure   SetDesignMode(nm: boolean);
   public
      Pic         : TImage;
      GridX       : integer;
      GridY       : integer;
      property    DrawPic: boolean read FDrawPic write SetDrawPic;
      property    CellBorder: integer read FCellBord write SetCellBorder;
      property    DesignMode: boolean read FDMode write SetDesignMode;
   end;

   TJumpLabel = class(TLabel)
   protected
      FCanClick   : boolean;
      procedure   SetCanClick(b: boolean);
   public
      JumpDest    : string;
      JumpOut     : boolean;
      constructor Create(AOwner: TComponent); override;
      property    CanClick: boolean read FCanClick write SetCanClick;
   end;


   procedure Register;

   procedure RegisterWABDComponent(WABDClass: TWABD_ObjectClass);
   procedure RegisterWABDComponents(WABDClasses: array of TWABD_ObjectClass);
   function  GetWABDClassListCount: integer;
   function  GetWABDClassListItem(i: integer): TWABD_ObjectClass;

   function  HTML_To_ASCII(const Input: string): string;
   function  ASCII_To_HTML(const Input: string): string;
   function  GetLocalBrowserForm: TWABD_Form;

   function FindRepAll(var s: string; const FindStr, RepStr: string): boolean;


const
   CR                = #13#10;
   PIXELS_PER_CHAR_X = 6;                    // I just picked this number
   PIXELS_PER_CHAR_Y = 18;                   // I just picked this number
   WABD_VERSION      = 1.24;
   WABD_SES_ID_STR   = 'SESSIONID';          // Hidden Text field to hold session information
   WABD_STATUS_STR   = 'WABDSTATUS';         // Magic word to get status info remotely (eg. WABDSTATUS=1)
   WABD_LABEL_STR    = 'WABDLABEL';
   WABD_BUTTON_STR   = 'WABD_BUTTON_NAME';   // "Name" used for buttons
   WABD_RADIO_STR    = 'WABD_RADIO_NAME';    // "Name" used for radiobuttons
   BACKGND_PIC_NAME  = 'FormBgndImg';        // Used w/ LocalBrowser and Form editor
   LabPointSizes     : array[1..7] of integer = (6,8,9,12,18,24,36);


var
   // These are to interface with the Localbrowser
   LocalBrowserException : string;
   LocalBrowserLogoff    : boolean;
   WABDClassList         : TList;
   LiveImageNum          : integer;


implementation

{$IFDEF VER100}
   uses DsgnIntf, WABD_ISAPI, HTML_Input, TraceUnit, JPEG;
   // {$R WABD_Objects.dcr}
{$ELSE}
   uses DsgnIntf, WABD_ISAPI, HTML_Input, TraceUnit;
{$ENDIF}


const
{$IFDEF WABD_REGISTERED}
   WABD_VERSION_STR  = 'Web Application Builder for Delphi 1.24';
{$ELSE}
   WABD_VERSION_STR  = 'Web Application Builder for Delphi 1.24 (Unregistered)';
{$ENDIF}

var
   LocalBrowserForm  : TWABD_Form;


// ************************************************************************
//    I N T E R F A C E   C U T - O F F
// ************************************************************************

// ************************************************************************
// Helper Objects
// ************************************************************************

type
   TTableCell = record
      SObj  : TWABD_SectionObject;
      Skip  : boolean;
      SpanX : integer;
      SpanY : integer;
   end;
   PTableCell = ^TTableCell;
   TTableCellArray = array[0..0] of TTableCell;
   PTableCellArray = ^TTableCellArray;

   TTableGrid = class
   protected
      mx, my      : integer;
      FData       : PTableCellArray;
   public
      constructor Create;
      destructor  Destroy; override;
      procedure   SetSize(x,y: integer);
      function    GetCell(x,y: integer): PTableCell;
   end;

// ************************************************************************
// TJumpLabel
// ************************************************************************

constructor TJumpLabel.Create(AOwner: TComponent);
begin
   inherited;
end;

procedure TJumpLabel.SetCanClick(b: boolean);
begin
   FCanClick := b;
   if b then begin
   {$IFDEF VER100}
      Cursor := crHandPoint;
   {$ENDIF}
      Font.Style := Font.Style + [fsUnderline];
   end else begin
      Cursor := crDefault;
   end;
end;



// ************************************************************************
// TTableGrid
// ************************************************************************

constructor TTableGrid.Create;
begin
   FData := nil;
   mx := 0;
   my := 0;
end;

destructor TTableGrid.Destroy;
begin
   if FData<>nil then FreeMem(FData);
end;

procedure TTableGrid.SetSize(x,y: integer);
begin
   if FData<>nil then FreeMem(FData);
   GetMem(FData, x*y*sizeof(TTableCell));
   FillChar(FData^, x*y*sizeof(TTableCell), 0);
   mx := x;
   my := y;
end;

function TTableGrid.GetCell(x,y: integer): PTableCell;
begin
   {$R-}
   Assert(x < mx, 'GetCell: X out of bounds');
   Assert(y < my, 'GetCell: Y out of bounds');
   Result := @FData^[x * my + y];
end;


// ************************************************************************
// TWABD_Table_Strings
// ************************************************************************

constructor TWABD_Table_Strings.Create;
begin
   FData := nil;
   XSize := 0;
   YSize := 0;
end;

destructor TWABD_Table_Strings.Destroy;
begin
   FreeData;
   inherited;
end;

procedure TWABD_Table_Strings.SafeSetSize(x, y: integer);
var
   MinX, MinY  : integer;
   NewData     : PStringArray;
   xp, yp      : integer;
begin
   if XSize < x then MinX := XSize else MinX := x;
   if YSize < y then MinY := YSize else MinY := y;

   if (x = 0) or (y = 0) then begin
      NewData := nil;
   end else begin
      GetMem(NewData, x * y * sizeof(string));
      Initialize(NewData^, x * y);
   end;
   for xp := 0 to MinX-1 do
      for yp := 0 to MinY-1 do
         NewData^[xp * y + yp] := FData^[xp * YSize + yp];

   FreeData;
   XSize := x;
   YSize := y;
   FData := NewData;
end;

procedure TWABD_Table_Strings.SetSize(x,y: integer);
begin
   if (x=XSize) and (y=YSize) then exit;
   FreeData;
   XSize := x;
   YSize := y;
   if (XSize = 0) or (YSize = 0) then begin
      FData := nil;
   end else begin
      GetMem(FData, XSize * YSize * sizeof(string));
      Initialize(FData^, XSize * YSize);
   end;
end;

procedure TWABD_Table_Strings.FreeData;
begin
   if FData<>nil then begin
      Finalize(FData^, XSize * YSize);
      FreeMem(FData);
   end;
   FData := nil;
end;

function TWABD_Table_Strings.GetString(x,y: integer): string;
begin
   Assert((x < XSize) and (x >= 0), 'GetString: X out of Bounds');
   Assert((y < YSize) and (y >= 0), 'GetString: Y out of Bounds');

   Result := FData^[x * YSize + y];
end;

procedure TWABD_Table_Strings.SetString(x,y: integer; NewString: string);
begin
   Assert((x < XSize) and (x >= 0), 'SetString: X out of Bounds');
   Assert((y < YSize) and (y >= 0), 'SetString: Y out of Bounds');

   FData^[x * YSize + y] := NewString;
end;

procedure TWABD_Table_Strings.DefineProperties(Filer: TFiler);
begin
   Filer.DefineProperty('TableStrings', ReadProps, WriteProps, True);
end;

procedure TWABD_Table_Strings.WriteProps(Writer: TWriter);
var
   x, y : integer;
begin
   Writer.WriteListBegin;
   Writer.WriteInteger(XSize);
   Writer.WriteInteger(YSize);
   for x := 0 to XSize-1 do
      for y := 0 to YSize-1 do
         Writer.WriteString(Strings[x,y]);
   Writer.WriteListEnd;
end;

procedure TWABD_Table_Strings.ReadProps(Reader: TReader);
var
   x, y : integer;
begin
   Reader.ReadListBegin;
   x := Reader.ReadInteger;
   y := Reader.ReadInteger;
   SetSize(x,y);
   for x := 0 to XSize-1 do
      for y := 0 to YSize-1 do
         Strings[x,y] := Reader.ReadString;
   Reader.ReadListEnd;
end;

procedure TWABD_Table_Strings.Assign(Source: TPersistent);
var
   t     : TWABD_Table_Strings;
   x, y  : integer;
begin
   if Source=Self then exit;
   t := Source as TWABD_Table_Strings;
   SetSize(t.XSize, t.YSize);
   for x := 0 to XSize-1 do
      for y := 0 to YSize-1 do
         Strings[x,y] := t.Strings[x,y];
end;


// ************************************************************************
// TPaintPanel
// ************************************************************************

procedure TPaintPanel.Paint;
var
   x, y     : integer;
   mx, my   : integer;
   ox, oy   : integer;
begin
   inherited;
   if FDMode then begin
      for x := 0 to Width div GridX do
         for y := 0 to Height div GridY do
            Canvas.Pixels[x * GridX, y * GridY] := clBlack;
   end;

   if FDrawPic and not FDMode then begin
      mx := Width div Pic.Width + 1;
      my := Height div Pic.Height + 1;

      ox := Left mod Pic.Width;     // To line up with everyone else
      oy := Top mod Pic.Height;

      for x := 0 to mx do begin
         for y := 0 to my do begin
            Canvas.Draw(x * Pic.Width - ox, y * Pic.Height - oy, Pic.Picture.Graphic);
         end;
      end;
   end;

   if FCellBord > 0 then begin

      if FFormSec<>nil then begin
         for x := 0 to FFormSec.NumCol do begin
            Canvas.MoveTo(FFormSec.ColTot[x], 0);
            Canvas.LineTo(FFormSec.ColTot[x], Height);
         end;
         for y := 0 to FFormSec.NumRow do begin
            Canvas.MoveTo(0, FFormSec.RowTot[y]);
            Canvas.LineTo(Width, FFormSec.RowTot[y]);
         end;
      end else begin
         for x := 0 to Width div GridX do begin
            Canvas.MoveTo(x * GridX, 0);
            Canvas.LineTo(x * GridX, Height);
         end;
         for y := 0 to Height div GridY do begin
            Canvas.MoveTo(0, y * GridY);
            Canvas.LineTo(Width, y * GridY);
         end;
      end;

   end;
end;

procedure TPaintPanel.SetDrawPic(b: boolean);
begin
   FDrawPic := b;
   Invalidate;
end;

procedure TPaintPanel.SetCellBorder(nb: integer);
begin
   FCellBord := nb;
   Invalidate;
end;

procedure TPaintPanel.SetDesignMode(nm: boolean);
begin
   FDMode := nm;
   Invalidate;
end;


// ************************************************************************
// TWABD_SessionMgr
// ************************************************************************

constructor TWABD_SessionMgr.Create(AOwner: TComponent);
begin
   inherited;
   CreateTime  := Now;
   SessionList := TThreadList.Create;
   FGarbage    := True;
   FCheck      := 60;                     // Run the check every 60 seconds
   Register_WABD_Callback(OnFormSubmit);  // Can only be one Session Mgr at a time
   LogStrings  := TStringList.Create;
   InitializeCriticalSection(LogCS);
   InitializeCriticalSection(SesMgrCS);

   StopEvent   := CreateEvent(nil, False, False, nil);
   GarbageThrd := TGarbageThread.Create(True);
   GarbageThrd.SesMgr := self;
   GarbageThrd.FreeOnTerminate := False;
   GarbageThrd.Resume;
end;

destructor TWABD_SessionMgr.Destroy;
begin
   SetEvent(StopEvent);
   TRACE0('Waiting for Garbage Thread');
   GarbageThrd.WaitFor;
   TRACE0('Garbage Thread terminated');
   CloseHandle(StopEvent);

   SessionList.Free;
   GarbageThrd.Free;
   LogStrings.Free;
   DeleteCriticalSection(LogCS);
   DeleteCriticalSection(SesMgrCS);
   inherited;
end;

procedure TWABD_SessionMgr.DoDestroySession(Ses: TWABD_Session);
begin
   if not Assigned(OnDestroySession) then
      raise Exception.CreateFmt('DestroySession not defined for %s', [Name]);

   // Trace0('Before Destroy');
   EnterCriticalSection(LogCS);
   try
      OnDestroySession(Ses);
   finally
      LeaveCriticalSection(LogCS);
   end;
   // Trace0('After Destroy');
end;

procedure TWABD_SessionMgr.CheckLogOff(Ses: TWABD_Session);
begin
   if Ses.DidLogOff then begin
      if Assigned(Ses.OnLogOff) then Ses.OnLogOff(Self);
      SessionList.Remove(Ses);
      AddToLog(Format('Session %0.4d logged off', [Ses.SessionID]));

      DoDestroySession(Ses);

      LocalBrowserLogoff := True;
   end;
end;

function TWABD_SessionMgr.CreateNewSession: string;
var
   NewSes : TWABD_Session;
begin
   if not Assigned(OnCreateSession) then
      raise Exception.CreateFmt('CreateSession not defined for %s', [Name]);

   NewSes := nil;
   EnterCriticalSection(LogCS);
   try
      OnCreateSession(NewSes);
   finally
      LeaveCriticalSection(LogCS);
   end;
   if NewSes=nil then
      raise Exception.CreateFmt('CreateSession Failed (NewSes = nil) for %s', [Name]);

   SessionList.Add(NewSes);

   Inc(NewSesID);
   NewSes.FSessionID := NewSesID;
   NewSes.LastAccess := Now;
   NewSes.CreateTime := Now;
   NewSes.FDLLName   := DLLName;

   Result := NewSes.ProcessInput(nil);
   CheckLogOff(NewSes);
end;

function TWABD_SessionMgr.RunExistingSession(const IdStr: string; InputList: TStrings): string;
var
   sList    : TList;
   i        : integer;
   ID       : integer;
   Ses, ts  : TWABD_Session;
begin
   // We have an existing session
   sList := SessionList.LockList;
   ID := StrToInt(IdStr);
   Ses := nil;
   for i := 0 to sList.Count-1 do begin
      ts := TWABD_Session(sList.Items[i]);
      if ts.SessionID = ID then begin
         Ses := ts;
         Ses.LastAccess := Now;
      end;
   end;
   SessionList.UnlockList;

   if Ses=nil then raise Exception.CreateFmt('Session ID not Found:  %s  (Session may have timed out', [IdStr]);

   AddToLog(Format('Session %0.4d = %s', [Ses.SessionID, EncodeEscapes(Copy(InputList.Text, 1, 40))]));
   Result := Ses.ProcessInput(InputList);
   CheckLogOff(Ses);
end;

procedure TWABD_SessionMgr.AddToLog(const s: string);
var
   dt : string;
begin
   dt := FormatDateTime('mmm dd yy hh:nn:ss', Now);

   EnterCriticalSection(LogCS);
   try
      LogStrings.Add(dt + ' : ' + s);
   finally
      LeaveCriticalSection(LogCS);
   end;
end;

procedure TWABD_SessionMgr.LogUserInfo(var ECB: TEXTENSION_CONTROL_BLOCK);
var
   RemoteAddr  : string;
   RemoteHost  : string;
   RemoteUser  : string;
   s           : string;
begin
   RemoteAddr := ISAPI_GetOneVar(ECB, 'REMOTE_ADDR');
   RemoteHost := ISAPI_GetOneVar(ECB, 'REMOTE_HOST');
   RemoteUser := ISAPI_GetOneVar(ECB, 'REMOTE_USER');
   s := Format('Created Session %0.4d  (%-15s %-15s %-15s)', [NewSesID+1, RemoteAddr, RemoteHost, RemoteUser]);
   AddToLog(s);
end;

function TWABD_SessionMgr.FormatStatus: string;
const
   NP = ' &nbsp <P>'#13#10;
var
   i : integer;
begin
   Result := '<HTML><BODY><CODE>' + NP;

   Result := Result + WABD_VERSION_STR + NP + NP;

   Result := Result + Format('Current Time: %s', [FormatDateTime('mmm dd yyyy hh:nn:ss', Now)]) + NP;
   Result := Result + Format('  Created On: %s', [FormatDateTime('mmm dd yyyy hh:nn:ss', CreateTime)]) + NP;
   Result := Result + Format('Elapsed Time: %s', [FormatDateTime('hh:nn:ss', Now-CreateTime)]) + NP + NP;

   for i := 0 to LogStrings.Count-1 do begin
      Result := Result + Format('%0.4d  %s', [i+1, LogStrings[i]]) + NP;
   end;
   Result := Result + NP + '</CODE></BODY></HTML>';
end;

function TWABD_SessionMgr.OnFormSubmit(var ECB: TEXTENSION_CONTROL_BLOCK): string;
var
   Input       : string;
   IdStr       : string;
   StatStr     : string;
   InputList   : TStringList;
begin
   try
      DLLName := ISAPI_GetOneVar(ECB, 'SCRIPT_NAME');

      InputList := TStringList.Create;
      GetInputString(ECB, Input);
      ParseHTMLInput(Input, InputList);

      StatStr := InputList.Values[WABD_STATUS_STR];
      if StatStr<>'' then begin
         Result := FormatStatus;
      end else begin
         IdStr := InputList.Values[WABD_SES_ID_STR];
         if IdStr<>'' then begin
            Result := RunExistingSession(IdStr, InputList);
         end else begin
            LogUserInfo(ECB);
            Result := CreateNewSession;
         end;
      end;

      InputList.Free;
      LocalBrowserException := '';
   except
      on e: Exception do begin
         LocalBrowserException := e.Message;
         AddToLog('EXCEPTION: ' + LocalBrowserException);
         Result := '<HTML><BODY>Error:  ' + LocalBrowserException + '</BODY></HTML>';
      end;
   end;
end;

procedure TGarbageThread.DoGarbageCollection;
var
   sList : TList;
   i     : integer;
   s     : TWABD_Session;
   Elap  : integer;
begin
   if SesMgr.GarbageCollection = False then exit;

   sList := SesMgr.SessionList.LockList;
   for i := sList.Count-1 downto 0 do begin
      s := TWABD_Session(sList.Items[i]);
      Elap := Round((Now - s.LastAccess) * 24.0 * 3600.0);
      if Elap > s.TimeOutLength then begin
         if Assigned(s.OnTimeOut) then s.OnTimeOut(Self);
         SesMgr.AddToLog(Format('TIME OUT: Session %0.4d', [s.SessionID]));

         SesMgr.DoDestroySession(s);

         sList.Delete(i);
      end;
   end;

   SesMgr.SessionList.UnlockList;
end;

procedure TGarbageThread.Execute;
var
   rc : integer;
begin
   repeat
      rc := WaitForSingleObject(SesMgr.StopEvent, SesMgr.FCheck * 1000);
      if rc = WAIT_TIMEOUT	then DoGarbageCollection;
   until rc <> WAIT_TIMEOUT;
end;


// ************************************************************************
// TWABD_Session
// ************************************************************************

constructor TWABD_Session.Create(AOwner: TComponent);
begin
   inherited;
   TimeOutLength := 600;
end;

destructor TWABD_Session.Destroy;
begin
   inherited;
end;

procedure TWABD_Session.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited;
   if (Operation = opRemove) and (AComponent = MainForm) then
      FMainForm := nil;
end;

function TWABD_Session.ProcessInput(InputList: TStrings): string;
begin
   if InputList = nil then begin
      // This is a new logon
      if not Assigned(MainForm) then
         raise Exception.CreateFmt('MainForm not defined for Session: %s', [Name]);
      MainForm.FWSession := self;
      NewForm := MainForm;
      if Assigned(OnFirstLogon) then OnFirstLogon(Self); // NewForm might change
      FCurForm := NewForm;
      CurForm.FWSession := self;
      if Assigned(CurForm.OnShow) then CurForm.OnShow(Self);
      FCurForm := NewForm;
      CurForm.FWSession := self;
      Result := CurForm.Object_To_HTML;
   end else begin
      // We have an existing form
      NewForm := CurForm;
      CurForm.FWSession := self;

      LocalError := '';
      try
         CurForm.SetProperties(InputList);
         CurForm.Call_Handler(InputList); // NewForm might change
      except
         on e: Exception do LocalError := e.Message;
      end;

      FCurForm := NewForm;
      CurForm.FWSession := self;
      Result := CurForm.Object_To_HTML;
   end;
   LastAccess := Now;

   LocalBrowserForm := CurForm;
end;

procedure TWABD_Session.LogOff;
begin
   DidLogOff := True;
end;

procedure TWABD_Session.Loaded;
var
   c : TComponent;
   i : integer;
begin
   inherited;
   Assert(Owner<>nil, Format('TWABD_Session %s has no owner', [Name]));
   for i := 0 to Owner.ComponentCount-1 do begin
      c := Owner.Components[i];
      if c is TWABD_Form then (c as TWABD_Form).Session := self;
   end;
end;


// ************************************************************************
// TWABD_Object
// ************************************************************************

constructor TWABD_Object.Create(AOwner: TComponent);
begin
   inherited;
   FVisible := True;
   FOrder   := -1;
   InLoaded := False;
end;

destructor TWABD_Object.Destroy;
begin
   // TRACE('DESTROYING Object:  %s (%s)', [Name, ClassName]);
   if FParent<>nil then begin
      FParent.FWABD_Objs.Remove(Self);
   end;

   inherited;
end;

procedure TWABD_Object.SetParent(NewParent: TWABD_Parent);
begin
   if NewParent=FParent then exit;

   if FParent<>nil then begin
      Assert(FParent.FWABD_Objs<>nil, 'FParent.WABD_Objs = nil');
      FParent.FWABD_Objs.Remove(Self);
   end;

   FParent := NewParent;
   if FParent<>nil then begin
      if (FOrder<>-1) and (InLoaded) then begin
         Assert(FParent.FWABD_Objs<>nil, 'FParent.WABD_Objs = nil');
         if FParent.FWABD_Objs.Count < Forder+1 then FParent.FWABD_Objs.Count := FOrder+1;
         FParent.FWABD_Objs.Items[FOrder] :=  Self;
      end else begin
         FParent.FWABD_Objs.Add(Self);
      end;
   end;
end;

function TWABD_Object.GetOrder: integer;
begin
  if FParent <> nil then begin
    Assert(FParent.FWABD_Objs<>nil, 'FParent.WABD_Objs = nil');
    Result := FParent.FWABD_Objs.IndexOf(Self);
   end else begin
    Result := -1;
   end;
end;

procedure TWABD_Object.SetOrder(NewOrder: integer);
var
   Count    : integer;
   CurOrder : integer;
begin
   if csLoading in ComponentState then begin
      FOrder := NewOrder;
   end else begin
      CurOrder := GetOrder;
      if CurOrder >= 0 then begin
         Assert(FParent<>nil, 'FParent = nil');
         Count := FParent.FWABD_Objs.Count;
         if NewOrder < 0 then NewOrder := 0;
         if NewOrder >= Count then NewOrder := Count - 1;
         if NewOrder <> CurOrder then begin
            Assert(FParent.FWABD_Objs<>nil, 'FParent.WABD_Objs = nil');
            FParent.FWABD_Objs.Delete(CurOrder);
            FParent.FWABD_Objs.Insert(NewOrder, Self);
         end;
      end;
   end;
end;

procedure TWABD_Object.SetVisible(b: boolean);
begin
   if b<>FVisible then begin
      FVisible := b;
      Changed;
   end;
end;

procedure TWABD_Object.SetName(const Value: TComponentName);
var
   OldName, NewName  : string;
   ParForm           : TWABD_Form;
begin
   OldName := Name;
   NewName := Value;
   inherited;
   Changed;

   if (not (csLoading in ComponentState)) then begin
      ParForm := GetParentForm;
      if ParForm<>nil then begin
         if Assigned(ParForm.OnChildNameChanged) then
            ParForm.OnChildNameChanged(Self, OldName, NewName);
      end;
   end;
end;

procedure TWABD_Object.Changed;
begin
   if Assigned(OnChange) then OnChange(Self);
end;

function TWABD_Object.GetParentForm: TWABD_Form;
var
   tmp : TWABD_Parent;
begin
   tmp := Parent;
   while (tmp<>nil) and (not (tmp is TWABD_Form)) do tmp := tmp.Parent;
   Result := TWABD_Form(tmp);
end;

procedure TWABD_Object.DefineProperties(Filer: TFiler);
begin
   inherited;
   Filer.DefineProperty('ParentAndOrder', ReadParentName, WriteParentName, True);
end;

procedure TWABD_Object.ReadParentName(Reader: TReader);
begin
   Reader.ReadListBegin;
   FParentName := Reader.ReadString;
   Order := Reader.ReadInteger;
   Reader.ReadListEnd;
end;

procedure TWABD_Object.WriteParentName(Writer: TWriter);
begin
   Writer.WriteListBegin;
   if Parent<>nil then
      Writer.WriteString(Parent.Name)
   else
      Writer.WriteString('');
   Writer.WriteInteger(Order);
   Writer.WriteListEnd;
end;

procedure TWABD_Object.Loaded;
begin
   inherited;
   InLoaded := True;
   Assert(Owner<>nil, 'Owner = nil');
   Parent := Owner.FindComponent(FParentName) as TWABD_Parent;
   InLoaded := False;
end;


// ************************************************************************
// TWABD_Parent
// ************************************************************************

constructor TWABD_Parent.Create(AOwner: TComponent);
begin
   inherited;
   FWABD_Objs := TList.Create;
end;

destructor TWABD_Parent.Destroy;
var
   i     : integer;
   Child : TWABD_Object;
begin
   // TRACE('BEGIN DESTROYING Parent:  %s (%s)', [Name, ClassName]);
   // This loop MUST go backwards because the child objects remove
   // *themselves* from the list we are iterating
   for i := FWABD_Objs.Count-1 downto 0 do begin
      Child := FWABD_Objs[i];
      Child.FParent := nil;      // Must use the non-property Parent variable
      Child.Free;
   end;
   // TRACE('END DESTROYING Parent:  %s (%s)', [Name, ClassName]);
   // TRACE0('');

   FWABD_Objs.Free;
   inherited;
end;

function TWABD_Parent.GetWABDObjects(i: integer): TWABD_Object;
begin
   Assert(i < FWABD_Objs.Count, 'Children: i out of bounds');
   Result := FWABD_Objs.Items[i];
end;

function TWABD_Parent.GetWABDObjCount: integer;
begin
   Result := FWABD_Objs.Count;
end;

procedure TWABD_Parent.ChildChanged(Sender: TObject);
begin
   // Default is to do nothing
end;

function TWABD_Parent.ForEachChild(ForEachProc: TWABD_ForEach; UserData: pointer): boolean;
var
   Stop  : boolean;
   i     : integer;
   tmp   : TWABD_Object;
   par   : TWABD_Parent;
begin
   Stop := False;

   for i := 0 to ChildCount-1 do begin
      ForEachProc(Children[i], Stop, UserData);
      if Stop then begin
         Result := True;
         exit;
      end;
   end;

   // Search for "grandchildren" if necessary
   for i := 0 to ChildCount-1 do begin
      tmp := Children[i];
      if (tmp is TWABD_Parent) then begin
         par := tmp as TWABD_Parent;
         if par.ForEachChild(ForEachProc, UserData) then begin
            Result := True;
            exit;
         end;
      end;
   end;
   Result := False;
end;

procedure TWABD_Parent.ChildNameProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
begin
   if Child.Name = tmp then begin
      TheChild := Child;
      Stop := True;
   end;
end;

function TWABD_Parent.ChildByName(ChildName: string): TWABD_Object;
begin
   tmp := ChildName;
   TheChild := nil;
   ForEachChild(ChildNameProc, nil);
   Result := TheChild;
end;

function TWABD_Parent.GetDefaultButton: TWABD_Button;
begin
   TheChild := nil;
   ForEachChild(DefButProc, nil);
   Result := TheChild as TWABD_Button;
end;

procedure TWABD_Parent.DefButProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
begin
   if (Child is TWABD_Button) then
      if (Child as TWABD_Button).Default then begin
         TheChild := Child;
         Stop := True;
      end;
end;

procedure TWABD_Parent.ButCapProc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
begin
   if (Child is TWABD_Button) then
      if (Child as TWABD_Button).Caption = tmp then begin
         TheChild := Child;
         Stop := True;
      end;
end;

function TWABD_Parent.ButtonByCaption(Caption: string): TWABD_Button;
begin
   tmp := Caption;
   TheChild := nil;
   ForEachChild(ButCapProc, nil);
   Result := TheChild as TWABD_Button;
end;

procedure TWABD_Parent.Loaded;
begin
   inherited;
end;

function TWABD_Parent.Object_To_Control(AOwner: TWinControl): TControl;
var
   np       : TPaintPanel;
   i        : integer;
   c        : TWABD_Object;
   con      : TControl;
   MaxHgt   : integer;
   MaxWid   : integer;
   CurTop   : integer;
begin
   np := TPaintPanel.Create(AOwner);
   np.Parent  := AOwner;   // Some of the controls need an HWND to set their properties
   np.Visible := False;    // For Performance (no screen updates)
   np.Name    := Name;
   np.Caption := '';

   MaxHgt := 0;
   MaxWid := 0;
   CurTop := 0;
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_Object;
      if (c.Visible=False) and not (csDesigning in ComponentState) then continue;

      con := c.Object_To_Control(np);
      if con<>nil then begin
         con.Name    := c.Name;
         con.Parent  := np;
         con.Top     := CurTop;
         con.Left    := 0;
         CurTop      := CurTop + con.Height;
         MaxHgt      := MaxHgt + con.Height;
         if con.Width > MaxWid then
            MaxWid := con.Width;
      end;
   end;
   np.Height := MaxHgt;
   np.Width  := MaxWid;
   np.BevelOuter := bvNone;

   np.Visible := True;
   Result := np;
end;


// ************************************************************************
// TWABD_Form
// ************************************************************************

procedure TWABD_Form.SetEdForm(NewEdForm: TWABDEditForm);
begin
   // Do nothing
end;

constructor TWABD_Form.Create(AOwner: TComponent);
begin
   inherited;
   FTextColor  := clNone;
   FLinkColor  := clNone;
   FVLinkColor := clNone;
   FBgndColor  := clNone;
   FEdForm     := TWABDEditForm.Create;
   FEdForm.ParForm := self;

   if Assigned(OnCreate) then OnCreate(Self);
end;

destructor TWABD_Form.Destroy;
begin
   if Assigned(RefNotify) then RefNotify(Self, Self, opRemove);
   inherited;
end;

procedure TWABD_Form.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FormTitle='') or (FormTitle=Name)) then FormTitle := Value;
   inherited;
end;

procedure TWABD_Form.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited;
   if Assigned(RefNotify) then RefNotify(Self, AComponent, Operation);
end;

procedure TWABD_Form.UncheckRadioButs(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
begin
   if (Child is TWABD_RadioButton) then
      (Child as TWABD_RadioButton).Checked := False;
   if (Child is TWABD_CheckBox) then
      (Child as TWABD_CheckBox).Checked := False;
end;

procedure TWABD_Form.SetProperties(InputList: TStrings);
var
   i        : integer;
   OneLine  : string;
   NameVal  : string;
   tmp      : string;
   c        : TWABD_Object;
begin
   ForEachChild(UncheckRadioButs, nil);
   for i := 0 to InputList.Count-1 do begin
      OneLine := InputList.Strings[i];
      NameVal := InputList.Names[i];

      c := ChildByName(NameVal);
      if (c<>nil) and (c is TWABD_Table) then continue;

      if (NameVal = WABD_SES_ID_STR) or (NameVal = WABD_BUTTON_STR)
         or (NameVal = WABD_LABEL_STR) then continue;

      if (NameVal = WABD_RADIO_STR) then begin
         tmp := InputList.Values[NameVal];
         c := ChildByName(tmp);
         Assert(c<>nil, 'Radio Button does not exist! ' + tmp);
         (c as TWABD_RadioButton).Checked := True;
         continue;
      end;

      // Skip Image clicks
      if (Pos('.X', UpperCase(NameVal)) <> 0) or
         (Pos('.Y', UpperCase(NameVal)) <> 0) then continue;

      c := ChildByName(NameVal);
      Assert(c<>nil, Format('No property named: %s', [NameVal]));
      c.HTML_To_Object(HTML_To_ASCII(InputList.Values[NameVal]));
   end;
end;

procedure TWABD_Form.ParseImageParams(Inputs: TStrings; var ImageName: string; var x,y: integer);
var
   i, p  : integer;
   n     : string;
begin
   ImageName := '';
   x  := -1;
   y  := -1;

   for i := 0 to Inputs.Count-1 do begin
      n := Inputs.Names[i];
      p := Pos('.X', UpperCase(n));
      if p<>0 then begin
         ImageName := Copy(n, 1, p-1);
         x  := StrToInt(Inputs.Values[n]);
      end;
      p := Pos('.Y', UpperCase(n));
      if p<>0 then begin
         y  := StrToInt(Inputs.Values[n]);
      end;
   end;
end;

procedure TWABD_Form.Call_Handler(InputList: TStrings);
var
   i           : integer;
   OneLine     : string;
   NameVal     : string;
   c           : TWABD_Object;
   b           : TWABD_Button;
   wt          : TWABD_Table;
   l           : TWABD_Label;
   ImageName   : string;
   im          : TWABD_Base_Image;
   x, y        : integer;
   ct          : string;
begin
   ParseImageParams(InputList, ImageName, x, y);
   if ImageName<>'' then begin
      c := ChildByName(ImageName);
      im := c as TWABD_Base_Image;
      im.MouseDown(x, y);
      exit;
   end;

   for i := 0 to InputList.Count-1 do begin
      OneLine := InputList.Strings[i];
      NameVal := InputList.Names[i];
      if NameVal = WABD_SES_ID_STR then continue;

     if (NameVal = WABD_BUTTON_STR) then begin
         c := ButtonByCaption(InputList.Values[NameVal]);
         Assert(c<>nil, 'Button does not exist: ' + InputList.Values[NameVal]);
         b :=  c as TWABD_Button;
         if Assigned(b.OnClick) then b.OnClick(b);
         exit;
      end;

      if (NameVal = WABD_LABEL_STR) then begin
         c := ChildByName(InputList.Values[NameVal]);
         Assert(c<>nil, 'Label does not exist: ' + InputList.Values[NameVal]);
         l := c as TWABD_Label;
         if Assigned(l.OnClick) then l.OnClick(l);
         exit;
      end;

      c := ChildByName(NameVal);
      if (c<>nil) and (c is TWABD_Table) then begin
         wt := c as TWABD_Table;
         ct := InputList.Values[NameVal];
         ct := Copy(ct, Length(wt.ClickText)+1, Length(ct));
         if Assigned(wt.OnClick) then wt.OnClick(wt, StrToInt(ct));
         exit;
      end;
   end;

   // You can get here if the user just presses enter in an Edit box.
   b := GetDefaultButton;
   if b<>nil then begin
      if Assigned(b.OnClick) then b.OnClick(b);
      exit;
   end;

   // Not sure what to do here, maybe have a form OnClick event? -bpz
   raise Exception.Create('Could not find an event handler - No Default Button exists');
end;

procedure TWABD_Form.ChildChanged(Sender: TObject);
begin
end;

function TWABD_Form.GetFormBody: string;
var
   TextStr  : string;
   LinkStr  : string;
   VLinkStr : string;
   BGndStr  : string;
   BGColStr : string;
   function HTMLColStr(col: TColor; ColName: string): string;
   begin
      if col <> clNone then
         Result := Format(' %s=#%6.6x', [ColName, ColorToRGB(col)])
      else Result := '';
   end;
begin
   // <BODY TEXT="#00ffff" LINK="#ff00ff" VLINK="#800000" BACKGROUND="Image3.gif">

   TextStr  := HTMLColStr(FTextColor, 'TEXT');
   LinkStr  := HTMLColStr(FLinkColor, 'LINK');
   VLinkStr := HTMLColStr(FVLinkColor, 'VLINK');
   BGColStr := HTMLColStr(FBgndColor, 'BGCOLOR');

   if FBgrdImage<>nil then begin
      if RunningLocal then
         BGndStr := Format(' BACKGROUND=%s', ['file://'+FBgrdImage.LocalPath+FBgrdImage.ImagePath])
      else
         BGndStr := Format(' BACKGROUND=%s', [FBgrdImage.ImagePath]);
   end;

   Result := Format('<BODY%s%s%s%s%s>', [TextStr, LinkStr, VLinkStr, BGndStr, BGColStr]);
end;

function TWABD_Form.GetNagStr: string;
begin
   // if RunningLocal then begin Result := ''; exit; end;

   {$IFNDEF WABD_REGISTERED}
   Result :=
      '<H6>' +
      '** UNREGISTERED COPY** &nbsp; &nbsp; &nbsp;' +
      'Visit the <A HREF=http://www.radix.net/~bziegler/wabd/wabd.htm>' +
      'Web Application Builder for Delphi Home Page</A> for registration information.' +
      '</H6> <P>' + CR;
   {$ELSE}
   Result := '';
   {$ENDIF}
end;

function TWABD_Form.GetLocalError: string;
begin
   Result := '';
   if Session=nil then exit;
   if Session.LocalError<>'' then
      Result := '<H3> ERROR: ' + Session.LocalError + '<P>' + CR +
         'POSSIBLE CAUSE:  User hit the "Back" button in the browser <P></H3>' + CR ;
end;

function TWABD_Form.DoMain: string;
var
   i           : integer;
begin
   Result := '';
   for i := 0 to ChildCount-1 do begin
      if Children[i].Visible then
         Result := Result + Children[i].Object_To_HTML;
   end;
end;

function TWABD_Form.Object_To_HTML: string;
var
   Header   : string;
   Footer   : string;
   SesID    : string;
   ID       : integer;
   DLLName  : string;
begin
   if Session<>nil then begin
      DLLName := Session.DLLName;
      ID      := Session.SessionID;
   end else begin
      DLLName := 'Unknown_DLL';
      ID      := -1;
   end;

   // Add the Headers and Footers
   Header := '<HTML><HEAD>' + CR +
      Format('<META NAME="Generator" CONTENT="%s">', [WABD_VERSION_STR]) + CR +
      '<TITLE>' + FormTitle + '</TITLE></HEAD>' +
      GetFormBody + CR +
      '<FORM ACTION=' + DLLName + ' METHOD=POST>' + CR;
   Footer := '</FORM></BODY></HTML>' + CR;
      SesID  := Format('<input type=hidden name=%s value=%d>' + CR,
         [WABD_SES_ID_STR, ID]);

   Result := Header + SesID + GetNagStr + GetLocalError + DoMain + Footer;
end;

procedure TWABD_Form.HTML_To_Object(FormVal: string);
begin
end;

procedure TWABD_Form.Show;
begin
   if not Assigned(FWSession) then
      raise Exception.CreateFmt('TWABD_Form "%s" does not have a WABD_Session!', [Name]);
   FWSession.NewForm := Self;
   if Assigned(OnShow) then OnShow(Self);
end;


// ************************************************************************
// "TOP" Level Objects
// ************************************************************************


// TWABD_Header

constructor TWABD_Header.Create(AOwner: TComponent);
begin
   inherited;
   FNum := 1;
end;

procedure TWABD_Header.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FCaption='') or (FCaption=Name)) then Caption := Value;
   inherited;
end;

procedure TWABD_Header.SetNum(NewNum: integer);
begin
   if NewNum < 1 then NewNum := 1;
   if NewNum > 6 then NewNum := 6;
   FNum := NewNum;
end;

function TWABD_Header.Object_To_HTML: string;
begin
   Result := Format('<H%d>%s</H%d>'+CR, [HeaderNum, ASCII_To_HTML(Caption), HeaderNum]);
end;

procedure TWABD_Header.HTML_To_Object(FormVal: string);
begin
   Caption := FormVal; // HTML_To_ASCII(FormVal); // Already done before passed here
end;

function TWABD_Header.Object_To_Control(AOwner: TWinControl): TControl;
var
   nl : TLabel;
begin
   nl := TLabel.Create(AOwner);
   nl.Name        := Name;
   nl.Caption     := Caption;
   nl.Font.Name   := 'Arial';
   nl.Font.Size   := 9 + (6-HeaderNum) * 4;
   nl.Height      := Round(nl.Height * 1.2);
   nl.Transparent := True;
   Result := nl;
end;


// TWABD_Raw

constructor TWABD_Raw.Create(AOwner: TComponent);
begin
   inherited;
   FLines := TStringList.Create;
end;

destructor TWABD_Raw.Destroy;
begin
   FLines.Free;
   inherited;
end;

procedure TWABD_Raw.SetLines(NewLines: TStrings);
begin
   FLines.Assign(NewLines);
end;

procedure TWABD_Raw.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FLines.Text='') or (FLines.Text=Name)) then FLines.Text := Value;
   inherited;
end;

function TWABD_Raw.Object_To_HTML: string;
begin
   Result := Lines.Text;
end;

procedure TWABD_Raw.HTML_To_Object(FormVal: string);
begin
   Lines.Text := FormVal;
end;

function TWABD_Raw.Object_To_Control(AOwner: TWinControl): TControl;
begin
   Result := nil;
end;


// TWABD_FormSection_Base

constructor TWABD_FormSection_Base.Create(AOwner: TComponent);
begin
   inherited;
   FGridX      := 16;
   FGridY      := 16;
   FCellBorder := 0;
   FCellSpace  := 0;
   FCellPad    := 0;
end;

procedure TWABD_FormSection_Base.SetGridX(NewX: integer);
begin
   FGridX := NewX;
   Changed;
end;

procedure TWABD_FormSection_Base.SetGridY(NewY: integer);
begin
   FGridY := NewY;
   Changed;
end;

function TWABD_FormSection_Base.DoFormSection: string;
const
   MX = 128;
   MY = 128;
var
   i       : integer;
   x, y    : integer;
   x2,y2   : integer;
   tg      : TTableGrid;
   c       : TWABD_SectionObject;
   MaxX    : integer;
   MaxY    : integer;
   r       : string;
   ws      : string;
   cspan   : string;
   rspan   : string;
   p       : PTableCell;
   SetWid  : array[0..MX] of boolean;
   // NumSkip : integer;
begin
   tg := TTableGrid.Create;
   tg.SetSize(MX, MY);

   MaxX := -1;
   MaxY := -1;
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_SectionObject;
      if c.Visible=False then continue;
      x := c.LeftPos div GridX;
      y := c.TopPos  div GridY {+ 1};    // +1 to leave the "Width" cells in there
      if x < 0 then x := 0;
      if y < 0 then y := 0;
      if x >= MX then x := MX-1;
      if y >= MY then y := MY-1;
      p := tg.GetCell(x,y);
      p.SObj := c;
      if c.ColSpan <> -1 then p.SpanX := c.ColSpan else
         p.SpanX := c.Width div GridX + 1;
      if c.RowSpan <> -1 then p.SpanY := c.RowSpan else
         p.SpanY := c.Height div GridY + 1;

      if x+p.SpanX-1 > MaxX then MaxX := x+p.SpanX-1;
      if y+p.SpanY-1 > MaxY then MaxY := y+p.SpanY-1;
      for x2 := 0 to p.SpanX-1 do
         for y2 := 0 to p.SpanY-1 do begin
            tg.GetCell(x+x2,y+y2).Skip := True;
         end;
   end;

   // For second line and beyond, only output *necessary* columns
   // If there are no controls from that cell to the end of the columns,
   // make that cell the last one for this line.
   {for y := 1 to MaxY do begin
      x := MaxX;
      NumSkip := 0;
      while (x > 0) and (tg.GetCell(x,y).SObj = nil) do begin
         if tg.GetCell(x,y).Skip = False then NumSkip := NumSkip + 1;
         tg.GetCell(x,y).Skip := True;
         x := x - 1;
      end;
      tg.GetCell(x,y).SpanX := tg.GetCell(x,y).SpanX + NumSkip;
   end;}

   for x := 0 to MX do SetWid[x] := False;

   r := '';
   for y := 0 to MaxY do begin
      r := r + '<TR>' + CR;
      for x := 0 to MaxX do begin
         p := tg.GetCell(x,y);
         if (SetWid[x]=False) and (not p.Skip) then begin
            ws := Format(' WIDTH=%d', [GridX]);
            SetWid[x] := True;
         end else begin
            ws := '';
         end;
         if p.SObj<>nil then begin
            cspan := Format(' COLSPAN=%d', [p.SpanX]);
            if p.SpanX = 1 then cspan := '';
            rspan := Format(' ROWSPAN=%d', [p.SpanY]);
            if p.SpanY = 1 then rspan := '';

            r := r + Format('<TD%s%s%s>', [ws, cspan, rspan]);
            r := r + p.SObj.Object_To_HTML;
            r := r + '</TD>'+CR;
         end else begin
            if not p.Skip then begin
               r := r + Format('<TD%s>&nbsp</TD>'+CR, [ws]);
            end else begin
               // r := r + Format('<TD%s>Skip</TD>'+CR, [ws]);
            end;
         end;
      end; // x
      r := r + '</TR>'+CR;
   end; // y

   tg.Free;
   Result := r;
end;

function TWABD_FormSection_Base.Object_To_HTML: string;
var
   header    : string;
   footer    : string;
begin
   Result := '';

   header := Format('<TABLE BORDER=%d CELLSPACING=%d CELLPADDING=%d>'+ CR,
      [FCellBorder, FCellSpace, FCellPad]);
   footer := '</TABLE>' + CR;

   Result := header + DoFormSection + footer;
end;

procedure TWABD_FormSection_Base.HTML_To_Object(FormVal: string);
begin
   raise Exception.Create('WABD FormSections do not have Values!');
end;

procedure TWABD_FormSection_Base.AutoSizeRowCol;
var
   i, t     : integer;
   c        : TWABD_SectionObject;
   rs, cs   : integer;
begin
   if csDesigning in ComponentState then exit;

   // Place each control into a Cell (Row,Col) and set those protected properties
   // Determine # of Rows & Cols
   NumRow := -1;
   NumCol := -1;
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_SectionObject;
      c.Col := c.OrigLeft div GridX;
      Assert(c.Col < 255, Format('Max 255 Columns per FormSection: %d, %d, %d', [c.LeftPos, GridX, c.Col]));
      c.Row := c.OrigTop  div GridY;
      Assert(c.Row < 255, 'Max 255 Rows per FormSection');
      cs    := c.Width   div GridX;
      rs    := c.Height  div GridY;
      if c.Col+cs > NumCol then
         NumCol := c.Col+cs;
      if c.Row+rs > NumRow then
         NumRow := c.Row+rs;
   end;

   // Init RowSizes and ColSizes to GridX, GridY
   for i := 0 to NumCol-1 do
      ColSizes[i] := GridX;
   for i := 0 to NumRow-1 do
      RowSizes[i] := GridY;

   // Loop through each Col (& Row), and set its size to the max control size (+CellBorder) in the Col
   // Only do autosize for ColSpan & RowSpan = 1
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_SectionObject;
      // Check the ColSize
      t := c.Width + CellBorder * 2;
      if (t > ColSizes[c.Col]) and (c.ColSpan = 1) then
         ColSizes[c.Col] := t;
      // Check the RowSize
      t := c.Height + CellBorder * 2;
      if (t > RowSizes[c.Row]) and (c.RowSpan = 1) then
         RowSizes[c.Row] := t;
   end;

   // Calculate ColTot & RowTot
   ColTot[0] := 0;
   for i := 1 to NumCol do    // note that this goes 1 past the NumCol
      ColTot[i] := ColTot[i-1] + ColSizes[i-1];
   RowTot[0] := 0;
   for i := 1 to NumRow do
      RowTot[i] := RowTot[i-1] + RowSizes[i-1];

   // Set each control's LeftPos & TopPos to the new values (+CellBorder)
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_SectionObject;
      c.LeftPos := ColTot[c.Col] + CellBorder;
      c.TopPos  := RowTot[c.Row] + CellBorder;
   end;
end;

function TWABD_FormSection_Base.Object_To_Control(AOwner: TWinControl): TControl;
var
   np          : TPaintPanel;
   i           : integer;
   c           : TWABD_SectionObject;
   con         : TControl;
   MaxX, MaxY  : integer;
   x, y        : integer;
begin
   AutoSizeRowCol;

   np := TPaintPanel.Create(AOwner);

   np.FFormSec    := Self;
   np.Parent      := AOwner;
   np.Name        := Name;
   np.Caption     := '';
   np.GridX       := GridX;
   np.GridY       := GridY;
   np.CellBorder  := CellBorder;
   for i := 0 to ChildCount-1 do begin
      c := Children[i] as TWABD_SectionObject;
      if (c.Visible=False) and not (csDesigning in ComponentState) then continue;

      con := c.Object_To_Control(np);
      if con<>nil then begin
         con.Left    := c.LeftPos;
         con.Top     := c.TopPos;
         con.Width   := c.Width;
         con.Height  := c.Height;
         con.Name    := c.Name;
         con.Parent  := np;
         con.Tag     := i + 1;
      end;
   end;

   MaxX := -1;
   MaxY := -1;
   for i := 0 to np.ControlCount-1 do begin
      x := np.Controls[i].Left + np.Controls[i].Width;
      y := np.Controls[i].Top + np.Controls[i].Height;
      if x > MaxX then MaxX := x;
      if y > MaxY then MaxY := y;
   end;
   np.Width := MaxX + 2;
   np.Height := MaxY + 2;
   np.BevelOuter := bvNone;

   Result := np;
end;


function TWABD_FormSection_Base.AddControl(ControlClass: TWABD_SectionObjectClass; Col, Row: integer): TWABD_SectionObject;
begin
   Result := ControlClass.Create(Self);
   Result.Parent  := Self;
   Result.LeftPos := Col * GridX;
   Result.TopPos  := Row * GridY;
   Result.ColSpan := 1;
   Result.RowSpan := 1;
end;

procedure TWABD_FormSection_Base.ControlAtFunc(Child: TWABD_Object; var Stop: boolean; UserData: pointer);
var
   so      : TWABD_SectionObject;
begin
   if not (Child is TWABD_SectionObject) then exit;
   so := TWABD_SectionObject(Child);
   if (so.Col = FindCol) and (so.Row = FindRow) then begin
      FindCon := Child as TWABD_SectionObject;
      Stop := True;
   end;
end;

function TWABD_FormSection_Base.ControlAtPos(Col, Row: integer): TWABD_SectionObject;
begin
   FindCon := nil;
   ForEachChild(ControlAtFunc, FindCon);
   Result := FindCon;
end;


// TWABD_BaseTable

constructor TWABD_BaseTable.Create(AOwner: TComponent);
begin
   inherited;
   FClickText := 'Go';
   CanClick   := False;
end;


// TWABD_Table

constructor TWABD_Table.Create(AOwner: TComponent);
begin
   inherited;
   CellBorder  := 1;
   CellSpacing := 1;
   FixedRows   := 1;
   FStrings    := TWABD_Table_Strings.Create;
end;

destructor TWABD_Table.Destroy;
begin
   FStrings.Free;
   inherited;
end;

function TWABD_Table.GetBut(Row: integer): string;
begin
   Result := Format('<INPUT TYPE=SUBMIT NAME=%s VALUE="%s %d">',
      [Name, ClickText, Row]);
end;

function TWABD_Table.Object_To_HTML: string;
var
   r, c              : integer;
   DataOn, DataOff   : string;
   Data              : string;
begin
   Result := Format('<TABLE BORDER=%d>'+CR, [CellBorder]);

   for r := 0 to Cells.Rows-1 do begin
      Result := Result + '<TR>';
      if CanClick then begin
         if r > 0 then Data := GetBut(r) else Data := '&nbsp';
         Result := Result + '<TD>' + Data + '</TD>';
      end;

      for c := 0 to Cells.Cols-1 do begin
         if (r+1 <= FixedRows) or (c+1 <= FixedCols) then begin
            DataOn  := '<TD><B>';
            DataOff := '</B></TD>';
         end else begin
            DataOn  := '<TD>';
            DataOff := '</TD>';
         end;
         Data := Cells[c,r];
         if Data='' then Data := '&nbsp';
         Result := Result + DataOn + Data + DataOff;
      end;
      Result := Result + '</TR>' + CR;
   end;

   Result := Result + '</TABLE>'+CR;
end;

procedure TWABD_Table.HTML_To_Object(FormVal: string);
begin
   Assert(False, 'Tables can not set properties!');
end;

procedure TWABD_Table.SetStrings(NewStrings: TWABD_Table_Strings);
begin
   FStrings.Assign(NewStrings);
end;

function TWABD_Table.Object_To_Control(AOwner: TWinControl): TControl;
var
   nt        : TStringGridEx;
   x, y      : integer;
   cw, w, tw : integer;
begin
   nt := TStringGridEx.Create(AOwner);
   nt.ClickText := ClickText;
   nt.Name      := Name;
   nt.CanClick  := CanClick;

   nt.Options := nt.Options + [goColSizing, goRowSelect];
   nt.FixedCols := FixedCols;
   nt.FixedRows := FixedRows;
   nt.ColCount  := Cells.Cols;
   nt.RowCount  := Cells.Rows;
   tw := 0;
   for x := 0 to nt.ColCount-1 do begin
      cw := -1;
      for y := 0 to nt.RowCount-1 do begin
         nt.Cells[x,y] := Cells[x,y];
         w := Round(Length(Cells[x,y]) * PIXELS_PER_CHAR_X * 1.25);
         if w > cw then cw := w;
      end;
      nt.ColWidths[x] := cw;
      tw := tw + cw;
   end;
   nt.Width     := tw + 25;
   nt.Height    := nt.DefaultRowHeight * nt.RowCount + 25;
   Result := nt;
end;


// TWABD_DataTable

constructor TWABD_DataTable.Create(AOwner: TComponent);
begin
   inherited;
   FDataLink  := TDataLink.Create;
   FShowForm  := True;
   FNavButs   := True;
   FShowTable := True;
   FMaxRows   := 10;
   FNumCol    := 3;
   FColWidth  := 150;
   Stat       := '';
   DidAppend  := False;

   FFormSec := TWABD_FormSection.Create(Self);
   FFormSec.Parent := self;
   FNavForm := TWABD_FormSection.Create(Self);
   FNavForm.Parent := self;
   FTable := TWABD_Table.Create(Self);
   FTable.Parent := self;
   FTable.CanClick := True;
   FTable.OnClick := TableClick;
end;

procedure TWABD_DataTable.SetName(const NewName: TComponentName);
begin
   inherited;
   FTable.Name := Name + '_Tab';
end;

destructor  TWABD_DataTable.Destroy;
begin
   FDataLink.Free;
   inherited;
end;

procedure TWABD_DataTable.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited;
   if (Operation = opRemove) and (AComponent = FDataLink.DataSource) then
      FDataLink.DataSource := nil;
end;

function TWABD_DataTable.GetDataSource: TDataSource;
begin
   Result := FDataLink.DataSource;
end;

procedure TWABD_DataTable.SetDataSource(NewDataSource: TDataSource);
begin
   FDataLink.DataSource := NewDataSource;
   if NewDataSource<>nil then NewDataSource.FreeNotification(self);
end;

function TWABD_DataTable.Object_To_HTML: string;
var
   t1, t2, t3 : string;
begin
   Result := '';
   if DataSource = nil then exit;

   if ShowEditForm then begin
      InitForm;
      t1 := FFormSec.Object_To_HTML;
   end;

   if ShowNavButs then begin
      InitNavButs;
      t2 := FNavForm.Object_To_HTML;
   end;

   if ShowTable then begin
      InitTable;
      t3 := FTable.Object_To_HTML;
   end;

   Result := t1 + t2 + t3;
end;

procedure TWABD_DataTable.HTML_To_Object(FormVal: string);
begin
   Assert(False, 'Data Tables can not set properties!');
end;

function TWABD_DataTable.FieldTypeIsString(ft: TFieldType): boolean;
begin
   Result := True;
   case ft of
      ftUnknown, ftBytes, ftVarBytes, ftBlob,
      ftGraphic, ftFmtMemo, ftMemo, ftParadoxOle,
      ftDBaseOle, ftTypedBinary: Result := False;
   end;
end;

procedure TWABD_DataTable.FirstClick(Sender: TObject);
begin
   DidAppend := False;
   if DataSource.DataSet.BOF then Stat := 'Already at First Record'
      else Stat := 'Moved to First Record';
   DataSource.DataSet.First;
end;

procedure TWABD_DataTable.LastClick(Sender: TObject);
begin
   DidAppend := False;
   if DataSource.DataSet.EOF then Stat := 'Already at Last Record'
      else Stat := 'Moved to Last Record';
   DataSource.DataSet.Last;
end;

procedure TWABD_DataTable.NextClick(Sender: TObject);
begin
   DidAppend := False;
   DataSource.DataSet.Next;
   if DataSource.DataSet.EOF then Stat := 'Already at Last Record'
      else Stat := 'Moved to Next Record';
end;

procedure TWABD_DataTable.PrevClick(Sender: TObject);
begin
   DidAppend := False;
   DataSource.DataSet.Prior;
   if DataSource.DataSet.BOF then Stat := 'Already at First Record'
      else Stat := 'Moved to Previous Record';
end;

procedure TWABD_DataTable.NextPgClick(Sender: TObject);
begin
   DidAppend := False;
   DataSource.DataSet.MoveBy(MaxRows);
   Stat := 'Moved to Next Page';
end;

procedure TWABD_DataTable.PrevPgClick(Sender: TObject);
begin
   DidAppend := False;
   DataSource.DataSet.MoveBy(-MaxRows);
   Stat := 'Moved to Previous Page';
end;

procedure TWABD_DataTable.TableClick(Sender: TObject; RowIndex: integer);
begin
   DidAppend := False;
   DataSource.DataSet.MoveBy(RowIndex-1);
   Stat := 'Moved to Record';
end;

procedure TWABD_DataTable.AddClick(Sender: TObject);
begin
   DidAppend := False;
   try
      DataSource.DataSet.Append;
      DidAppend := True;
      Stat := 'Fields initialized, make changes and then select "Edit" to save';
      InitForm;
   except
      on e: Exception do begin
         Stat := e.Message;
         DataSource.DataSet.Cancel;
      end;
   end;
end;

procedure TWABD_DataTable.EditClick(Sender: TObject);
var
   i        : integer;
   we       : TWABD_Edit;
   EdName   : string;
   f        : TField;
   AsString : boolean;
begin
   try
      if DidAppend then
         DataSource.DataSet.Append
      else
         DataSource.DataSet.Edit;

      // Read in the Field Values
      with FFormSec do begin
         for i := 0 to DataSource.DataSet.FieldCount-1 do begin
            f := DataSource.DataSet.Fields[i];
            AsString := FieldTypeIsString(f.DataType);

            EdName := Self.Name + '_TFE_' + IntToStr(i);
            we := ChildByName(EdName) as TWABD_Edit;
            if AsString then begin
               f.AsString := we.Text;
            end;
         end;
      end;

      DataSource.DataSet.Post;
      Stat := 'Field updated';
      DidAppend := False;
   except
      on e: Exception do begin
         Stat := e.Message;
         DataSource.DataSet.Cancel;
         DidAppend := False;
      end;
   end;
end;

procedure TWABD_DataTable.DeleteClick(Sender: TObject);
begin
   DidAppend := False;
   try
      DataSource.DataSet.Delete;
      Stat := 'Deleted Record';
   except
      on e: Exception do Stat := e.Message;
   end;
end;

procedure TWABD_DataTable.InitForm;
var
   i     : integer;
   x, y  : integer;
   ds    : TDataSet;
   f     : TField;
   nl    : TWABD_Label;
   ne    : TWABD_Edit;
   AsString : boolean;
begin
   ds := DataSource.DataSet;
   with FFormSec do begin
      // Clear out old Controls
      for i := ChildCount-1 downto 0 do
         Children[i].Free;

      GridX := FColWidth;
      GridY := 25;

      x := 0;
      y := 0;
      for i := 0 to ds.FieldCount-1 do begin
         f := ds.Fields[i];

         // Create the Label
         nl := TWABD_Label.Create(Self);
         nl.Parent := FFormSec;
         nl.Caption := f.DisplayLabel;
         nl.LeftPos := x * GridX;
         nl.TopPos  := (y*2) * GridY;

         AsString := FieldTypeIsString(f.DataType);

         // Create the Edit Box
         ne := TWABD_Edit.Create(Self);
         ne.Name    := Self.Name + '_TFE_' + IntToStr(i);
         ne.Parent  := FFormSec;
         ne.LeftPos := x * GridX;
         ne.TopPos  := (y*2+1) * GridY;
         ne.Size    := (GridX - 10) div PIXELS_PER_CHAR_X;
         if AsString then
            ne.Text := f.AsString
         else
            ne.Text := '';

         // go to the next position
         x := x + 1;
         if x = FNumCol then begin
            x := 0;
            y := y + 1;
         end;
      end;
   end;
end;

procedure TWABD_DataTable.CreateNavBut(x, y : integer; ButCap: string; OnClick: TNotifyEvent);
var
   b : TWABD_Button;
begin
   b         := TWABD_Button.Create(Self);
   b.Parent  := FNavForm;
   b.LeftPos := x;
   b.TopPos  := y;
   b.Caption := ButCap;
   b.OnClick := OnClick;
end;

procedure TWABD_DataTable.InitNavButs;
const
   X = 50;
var
   i  : integer;
   nl : TWABD_Label;
begin
   // Clear out old Controls
   for i := FNavForm.ChildCount-1 downto 0 do
      FNavForm.Children[i].Free;

   FNavForm.GridX := X;
   FNavForm.GridY := 35;

   CreateNavBut(X * 0,  0, 'First',  FirstClick);
   CreateNavBut(X * 1,  0, 'Last',   LastClick);
   CreateNavBut(X * 2,  0, 'Next',   NextClick);
   CreateNavBut(X * 3,  0, 'Prev',   PrevClick);
   CreateNavBut(X * 4,  0, 'Next+',  NextPgClick);
   CreateNavBut(X * 5,  0, 'Prev+',  PrevPgClick);

   CreateNavBut(X * 7,  0, 'Add',    AddClick);
   CreateNavBut(X * 8,  0, 'Edit',   EditClick);
   CreateNavBut(X * 9,  0, 'Delete', DeleteClick);

   nl := TWABD_Label.Create(Self);
   nl.Parent := FNavForm;
   nl.LeftPos := 0;
   nl.TopPos  := 35;
   if Stat='' then Stat := 'Viewing ' + DataSource.DataSet.Name;
   nl.Caption := Format('Status:  %s', [Stat]);
end;

procedure TWABD_DataTable.InitTable;
var
   ds    : TDataSet;
   i     : integer;
   bm    : TBookmark;
   row   : integer;
   dt    : TFieldType;
begin
   ds := DataSource.DataSet;
   FTable.Cells.SetSize(0, 0);
   FTable.Cells.SetSize(ds.FieldCount, FMaxRows + 1);

   // Set the Field Headers
   for i := 0 to ds.FieldCount-1 do begin
      FTable.Cells[i, 0] := ds.Fields[i].DisplayLabel;
   end;

   bm := ds.GetBookmark;
   row := 0;
   repeat
      for i := 0 to ds.FieldCount-1 do begin
         if ds.Fields[i].Visible = False then continue;
         dt := ds.Fields[i].DataType;
         if not FieldTypeIsString(dt) then continue;
         FTable.Cells[i, row+1] := ds.Fields[i].AsString;
      end;
      ds.Next;
      row := row + 1;
   until (row >= FMaxRows) or (ds.EOF);

   ds.GotoBookmark(bm);
   ds.FreeBookmark(bm);
end;


// TWABD_Hidden

function TWABD_Hidden.Object_To_HTML: string;
begin
   Result := Format('<input type=hidden name=%s value=%d>' + CR,
      [Name, Value]);
end;

procedure TWABD_Hidden.HTML_To_Object(FormVal: string);
begin
   Value := FormVal;
end;

function TWABD_Hidden.Object_To_Control(AOwner: TWinControl): TControl;
var
   ne : TEdit;
begin
   ne := TEdit.Create(AOwner);
   ne.Parent   := AOwner;
   ne.Name     := Name;
   ne.Height   := 0;
   ne.Text     := Value;
   Result := ne;
end;


// TWABD_BlankLines

constructor TWABD_BlankLines.Create(AOwner: TComponent);
begin
   inherited;
   FNumLines := 1;
end;

function TWABD_BlankLines.Object_To_HTML: string;
var
   i : integer;
begin
   Result := '';
   for i := 0 to FNumLines-1 do begin
      Result := Result + '<P>&nbsp</P>';
   end;
end;

function TWABD_BlankLines.Object_To_Control(AOwner: TWinControl): TControl;
var
   np : TPaintPanel;
begin
   np := TPaintPanel.Create(AOwner);
   np.BevelOuter := bvNone;
   np.Height     := FNumLines * PIXELS_PER_CHAR_Y;
   np.Name       := Name;
   np.Caption    := '';
   Result := np;
end;

procedure TWABD_BlankLines.HTML_To_Object(FormVal: string);
begin
   Assert(False, 'Blank Lines can not have properties!');
end;


// ************************************************************************
// Form Level Objects
// ************************************************************************


// TWABD_SectionObject

constructor TWABD_SectionObject.Create(AOwner: TComponent);
begin
   inherited;
   FLeftPos  := 0;
   FTopPos   := 0;
   FWidth    := 0;
   FHeight   := 0;
   FColSpan  := -1;
   FRowSpan  := -1;
end;

procedure TWABD_SectionObject.SetLeft(NewLeft: integer);
begin
   if (NewLeft<>OrigLeft) then begin
      FLeftPos := NewLeft;
      OrigLeft := NewLeft;
      Changed;
   end;
end;

procedure TWABD_SectionObject.SetTop(NewTop: integer);
begin
   if (NewTop<>OrigTop) then begin
      FTopPos := NewTop;
      OrigTop := NewTop;
      Changed;
   end;
end;


// TWABD_Anchor

constructor TWABD_Anchor.Create(AOwner: TComponent);
begin
   inherited;
   Destination := 'http://www.radix.net/~bziegler/wabd/wabd.htm';
end;

function TWABD_Anchor.Object_To_HTML: string;
var
   r2 : string;
begin
   Result := Format('<A HREF=%s>', [Destination]);
   r2 := ASCII_To_HTML(Caption);
   if Bold      then r2 := '<B>' + r2 + '</B>';
   if Italic    then r2 := '<I>' + r2 + '</I>';
   if Underline then r2 := '<U>' + r2 + '</U>';
   Result := Result + r2 + '</A>';
end;

procedure TWABD_Anchor.HTML_To_Object(FormVal: string);
begin
   raise Exception.Create('WABD Anchors do not have Values!');
end;

procedure TWABD_Anchor.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FCaption='') or (FCaption=Name)) then Caption := Value;
   inherited;
end;

procedure TWABD_Anchor.SetCaption(NewCaption: string);
begin
   FCaption := NewCaption;
   FWidth := Length(Caption) * PIXELS_PER_CHAR_X;
   FHeight := PIXELS_PER_CHAR_Y + 1;
   Changed;
end;

function TWABD_Anchor.Object_To_Control(AOwner: TWinControl): TControl;
var
   nl : TJumpLabel;
begin
   nl := TJumpLabel.Create(AOwner);
   nl.CanClick    := True;
   nl.JumpDest    := FDest;
   {$IFDEF VER100}
   nl.Cursor      := crHandPoint;
   {$ENDIF}
   nl.AutoSize    := False;
   nl.Name        := Name;
   nl.Caption     := Caption;
   nl.Font.Style  := nl.Font.Style + [fsUnderLine];
   nl.Font.Size   := 9;
   nl.Font.Color  := clBlue;
   nl.Transparent := True;
   nl.JumpOut     := True;
   Result := nl;
end;


// TWABD_HotSpot

destructor TWABD_HotSpot.Destroy;
var
   OldChange : TNotifyEvent;
begin
   OldChange := FChange;
   inherited;
   if Assigned(OldChange) then OldChange(nil);
end;

procedure TWABD_HotSpot.SetX1(i: integer);
begin
   FX1 := i;
   Changed;
end;

procedure TWABD_HotSpot.SetY1(i: integer);
begin
   FY1 := i;
   Changed;
end;

procedure TWABD_HotSpot.SetX2(i: integer);
begin
   FX2 := i;
   Changed;
end;

procedure TWABD_HotSpot.SetY2(i: integer);
begin
   FY2 := i;
   Changed;
end;

procedure TWABD_HotSpot.SetName(const Value: TComponentName);
begin
   inherited;
end;

procedure TWABD_HotSpot.Changed; 
begin
   if Assigned(OnChange) then OnChange(Self);
end;


// TWABD_Base_Image

constructor TWABD_Base_Image.Create(AOwner: TComponent);
begin
   inherited;
   FHotSpots  := TWABD_HotSpots.Create;
   FHotSpots.ParImage := Self;
    
   FImgWidth  := 150;
   FImgHeight := 150;
   FWidth     := 150;
   FHeight    := 150;
   FLocalPath := 'c:\WebShare\WWWroot\';
end;

destructor  TWABD_Base_Image.Destroy;
begin
   inherited;
end;

procedure TWABD_Base_Image.SetHotSpots(HS: TWABD_HotSpots);
begin
   // Do Nothing
end;

function TWABD_Base_Image.Object_To_HTML: string;
var
   p : string;
begin
   if RunningLocal then begin
      p := 'file://' + LocalPath + ImagePath;
   end else begin
      p := ImagePath;
   end;
   Result := Format('<INPUT TYPE=IMAGE NAME=%s SRC="%s" ALT="%s">',
      [Name, p, ASCII_To_HTML(AltText)]);
end;

procedure TWABD_Base_Image.HTML_To_Object(FormVal: string);
begin
   raise Exception.Create('WABD Images do not have Values!');
end;

procedure TWABD_Base_Image.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FAltText='') or (FAltText=Name)) then FAltText := Value;
   inherited;
end;

procedure TWABD_Base_Image.SetImagePath(NewImagePath: string);
begin
   FImagePath := NewImagePath;
   if Copy(FImagePath, 1, 1)<>'\' then FImagePath := '\' + FImagePath;
   if not (csLoading in ComponentState) then UpdateImageSize;
end;

procedure TWABD_Base_Image.SetLocalPath(NewLocalPath: string);
begin
   FLocalPath := NewLocalPath;
   if Copy(FLocalPath, Length(FLocalPath), 1)='\' then
      FLocalPath := Copy(FLocalPath, 1, Length(FLocalPath)-1);
   if not (csLoading in ComponentState) then UpdateImageSize;
end;

procedure TWABD_Base_Image.Loaded;
begin
   inherited;
   UpdateImageSize;
end;

procedure TWABD_Base_Image.UpdateImageSize;
var
   f : string;
   p : TPicture;
begin
   if csDesigning in ComponentState then begin
      f := LocalPath + ImagePath;
      if FileExists(f) then begin
         p := TPicture.Create;
         try
            p.LoadFromFile(f);
            ImageWidth := p.Width;
            ImageHeight := p.Height;
         except
            on e: Exception do begin end;
         end;
         p.Free;
      end;
   end;
   Changed; 
end;

procedure TWABD_Base_Image.SetImgWidth(w: integer);
begin
   FImgWidth := w;
   FWidth := w;
   Changed;
end;

procedure TWABD_Base_Image.SetImgHeight(h: integer);
begin
   FImgHeight := h;
   FHeight := h;
   Changed;
end;

function TWABD_Base_Image.Object_To_Control(AOwner: TWinControl): TControl;
var
   ni : TImage;
   bm : Graphics.TBitmap;
begin
   ni := TImage.Create(AOwner);
   ni.Parent := AOwner;
   ni.Name   := Name;
   try
      ni.Picture.LoadFromFile(LocalPath + ImagePath);
   except
      on e: Exception do begin end;
   end;
   if ni.Picture.Width = 0 then begin
      bm := Graphics.TBitmap.Create;
      bm.Width := ImageWidth;
      bm.Height := ImageHeight;
      bm.Canvas.TextOut(0, 0, 'Unable to load Bitmap: ');
      bm.Canvas.TextOut(0, 25, LocalPath + ImagePath);
      ni.Picture.Bitmap := bm;
      bm.Free;
   end;
   Result := ni;
end;

procedure TWABD_Base_Image.MouseDown(x, y: integer);
var
   i  : integer;
   c  : TComponent;
   h  : TWABD_HotSpot;
   R  : TRect;
   pt : TPoint;
begin
   pt := Point(x, y);
   Assert(Owner<>nil, Format('Image %s has no Owner', [Name]));
   for i := 0 to Owner.ComponentCount-1 do begin
      c := Owner.Components[i];
      if (c is TWABD_HotSpot) then begin
         h := c as TWABD_HotSpot;
         R := Rect(h.X1, h.Y1, h.X2, h.Y2);
         if (h.ImageParent = Self) and PtInRect(R, pt) then begin
            if Assigned(h.OnClick) then h.OnClick(h);
            exit;
         end;
      end;
   end;

   if Assigned(OnMouseDown) then OnMouseDown(Self, x, y);
   if Assigned(OnClick) then OnClick(Self);
end;


// TWABD_LiveImage

constructor TWABD_LiveImage.Create(AOwner: TComponent);
begin
   inherited;
   FSafeBmp   := TSafeBitmap.Create(Self);
   FSafeBmp.PixelFormat := pf4bit;           // This is the safest format
   FDirty     := True;
   FCanvas    := nil;
   WroteFile  := False;
   FImgWidth  := 200;
   FWidth     := 200;
   FImgHeight := 100;
   FHeight    := 100;
end;

destructor TWABD_LiveImage.Destroy;
begin
   ReleaseCanvas;
   if WroteFile then DeleteFile(PChar(GetFileName));
   inherited;
end;

function TWABD_LiveImage.GetFileName: string;
begin
   Result := LocalPath + '\' + FFileName;
end;

procedure TWABD_LiveImage.SetImgWidth(w: integer);
begin
   inherited;
   if not (csLoading in ComponentState) then
      FSafeBmp.SetSize(ImageWidth, ImageHeight);
end;

procedure TWABD_LiveImage.SetImgHeight(h: integer);
begin
   inherited;
   if not (csLoading in ComponentState) then
      FSafeBmp.SetSize(ImageWidth, ImageHeight);
end;

function TWABD_LiveImage.GetSafeBitmap: TSafeBitmap;
begin
   FDirty := True;      // Assume it gets modified
   Result := FSafeBmp;
end;

function TWABD_LiveImage.GetCanvas: TCanvas;
begin
   ReleaseCanvas;       // Needed for Delphi 2 only?
   if FCanvas=nil then begin
      FCanvas := TCanvas.Create;
      FCanvas.Handle := SafeBitmap.GetDC;
   end;
   FDirty := True;
   Result := FCanvas;
end;

procedure TWABD_LiveImage.Loaded;
begin
   inherited;
   FSafeBmp.SetSize(ImageWidth, ImageHeight);
   FFileName := GetNewName;
   ImagePath := FFileName;
   Changed;
end;

function TWABD_LiveImage.GetNewName: string;
begin
   Inc(LiveImageNum);
   Result := Format('TMP%6.6x%4.4x.bmp', [Round(Random*$FFFFFF), LiveImageNum]);
end;

procedure TWABD_LiveImage.SaveImage;
begin
   if FDirty then begin
      if WroteFile then DeleteFile(PChar(GetFileName));
      WroteFile := False;
      FFileName := GetNewName;
      FImagePath := FFileName;
      if Copy(FImagePath, 1, 1)<>'\' then FImagePath := '\' + FImagePath;
      if FSafeBmp.SaveToFile(GetFileName) then WroteFile := True;
      FDirty := False;
   end;
end;

function TWABD_LiveImage.Object_To_HTML: string;
begin
   SaveImage;
   Result := inherited Object_To_HTML;
end;

procedure TWABD_LiveImage.ReleaseCanvas;
begin
   FCanvas.Free;
   FCanvas := nil;
end;

function TWABD_LiveImage.Object_To_Control(AOwner: TWinControl): TControl;
var
   ni : TImage;
   bm : Graphics.TBitmap;
begin
   SaveImage;
   ni := TImage.Create(AOwner);
   ni.Parent := AOwner;
   ni.Name   := Name;
   try
      ni.Picture.LoadFromFile(LocalPath + ImagePath);
   except
      on e: Exception do begin end;
   end;
   if ni.Picture.Width = 0 then begin
      bm := Graphics.TBitmap.Create;
      bm.Width := ImageWidth;
      bm.Height := ImageHeight;

      with bm.Canvas do begin
         Font.Name := 'Times New Roman';
         Font.Size := 24;
         Font.Style := Font.Style + [fsItalic, fsBold];
         Brush.Style := bsClear;

         Font.Color := clLtGray;
         bm.Canvas.TextOut(8, 8, 'Live Image');

         Font.Color := clBlue;
         bm.Canvas.TextOut(5, 5, 'Live Image');
      end;

      ni.Picture.Bitmap := bm;
      bm.Free;
   end;
   Result := ni;
end;


// TWABD_Label

constructor TWABD_Label.Create(AOwner: TComponent);
begin
   inherited;
   FFontSize   := 3;
   FFontColor  := clNone;
   FBold       := False;
   FItalic     := False;
   FUnderline  := False;
   FCanClick   := False;
end;

function TWABD_Label.GetHRef: string;
var
   pf : TWABD_Form;
begin
   pf := GetParentForm;
   if (pf=nil) or (pf.Session=nil) then begin
      Result := '';
   end else begin
      // Assert(GetParentForm <> nil);
      // Assert(GetParentForm.Session <> nil);
      Result := Format('%s?%s=%d&%s=%s', [pf.Session.DLLName,
         WABD_SES_ID_STR, pf.Session.SessionID, WABD_LABEL_STR, Name]);
   end;
end;

function TWABD_Label.Object_To_HTML: string;
var
   SizeStr  : string;
   ColStr   : string;
   rgb      : TRGBQuad;
   col      : integer;
begin
   Result := ASCII_To_HTML(Caption);
   if Bold      then Result := '<B>' + Result + '</B>';
   if Italic    then Result := '<I>' + Result + '</I>';
   if Underline then Result := '<U>' + Result + '</U>';
   if FontSize<>3 then SizeStr := ' SIZE='+IntToStr(FontSize);
   col := ColorToRGB(FontColor);
   Move(col, rgb, sizeof(rgb));
   if FontColor<>clNone then ColStr := ' COLOR=#'+Format('%2.2x%2.2x%2.2x',
      [rgb.rgbBlue, rgb.rgbGreen, rgb.rgbRed]);
   if (SizeStr<>'') or (ColStr<>'') then
      Result := '<FONT'+SizeStr+ColStr+'>' + Result + '</FONT>';

   if CanClick then begin
      Result := Format('<A HREF=%s>%s</A>', [GetHRef, Result]);
   end;
end;

procedure TWABD_Label.HTML_To_Object(FormVal: string);
begin
   raise Exception.Create('WABD Labels do not have Values!');
end;

procedure TWABD_Label.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FCaption='') or (FCaption=Name)) then Caption := Value;
   inherited;
end;

procedure TWABD_Label.SetBold(NewBold: boolean);
begin
   FBold := NewBold;
   UpdateWidHgt;
end;

procedure TWABD_Label.UpdateWidHgt;
var
   m : double;
begin
   FWidth := Length(FCaption) * (PIXELS_PER_CHAR_X + 1);
   FHeight := PIXELS_PER_CHAR_Y - 3;

   if FBold then m := 1.25 else m := 1.0;

   // Now update based on the FontSize
   FWidth  := Round(FWidth * LabPointSizes[FontSize] * m / 9.0);
   FHeight := Round(FHeight * LabPointSizes[FontSize] * m / 9.0);
end;

procedure TWABD_Label.SetCaption(NewCaption: string);
begin
   FCaption := NewCaption;
   UpdateWidHgt;
   Changed;
end;

function TWABD_Label.Object_To_Control(AOwner: TWinControl): TControl;
var
   nl : TJumpLabel;
begin
   nl := TJumpLabel.Create(AOwner);
   nl.AutoSize    := True;
   nl.Font.Size   := LabPointSizes[FontSize];
   nl.Font.Name   := 'Arial';
   if FontColor<>clNone then nl.Font.Color  := FontColor;
   if Bold then nl.Font.Style := nl.Font.Style + [fsBold];
   if Italic then nl.Font.Style := nl.Font.Style + [fsItalic];
   if Underline then nl.Font.Style := nl.Font.Style + [fsUnderline];
   nl.Name        := Name;
   nl.Caption     := Caption;
   nl.Transparent := True;
   nl.CanClick    := CanClick;
   nl.JumpDest    := GetHRef;
   nl.JumpOut     := False;
   Result := nl;
end;

procedure TWABD_Label.SetFontSize(NewSize: integer);
begin
   if NewSize < 1 then NewSize := 1;
   if NewSize > 7 then NewSize := 7;
   FFontSize := NewSize;
   Changed;
end;


// TWABD_LinesObject

constructor TWABD_LinesObject.Create(AOwner: TComponent);
begin
   inherited;
   FLines := TStringList.Create;
end;

destructor TWABD_LinesObject.Destroy;
begin
   FLines.Free;
   inherited;
end;

procedure TWABD_LinesObject.SetLines(NewLines: TStrings);
begin
   FLines.Assign(NewLines);
end;

procedure TWABD_LinesObject.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FLines.Text='') or (FLines.Text=Name)) then FLines.Text := Value;
   inherited;
end;

// TWABD_Memo

constructor TWABD_Memo.Create(AOwner: TComponent);
begin
   inherited;
   Rows := 5;
   Cols := 20;
end;

function TWABD_Memo.Object_To_HTML: string;
var
   i : integer;
begin
   Result := Format('<TEXTAREA NAME=%s COLS=%d ROWS=%d>', [Name, Cols, Rows]);
   for i := 0 to Lines.Count-1 do begin
      Result := Result + ASCII_To_HTML(Lines[i]) + CR;
   end;
   Result := Result + '</TEXTAREA>';
end;

procedure TWABD_Memo.HTML_To_Object(FormVal: string);
begin
   Lines.Text := FormVal;
end;

procedure TWABD_Memo.SetCols(NewCols: integer);
begin
   FCols := NewCols;
   FWidth := FCols * PIXELS_PER_CHAR_X;
   Changed;
end;

procedure TWABD_Memo.SetRows(NewRows: integer);
begin
   FRows := NewRows;
   FHeight := FRows * PIXELS_PER_CHAR_Y;
   Changed;
end;

function TWABD_Memo.Object_To_Control(AOwner: TWinControl): TControl;
var
   nm : TMemo;
begin
   nm := TMemo.Create(AOwner);
   nm.Parent      := AOwner;
   nm.Name        := Name;
   nm.Lines.Text  := Lines.Text;
   Result := nm;
end;


// TWABD_Edit

constructor TWABD_Edit.Create(AOwner: TComponent);
begin
   inherited;
   Size  := 10;
   FPass := False;
end;

function TWABD_Edit.Object_To_HTML: string;
var
   r : string;
begin
   if not IsPassword then
      r := '<INPUT TYPE=TEXT '
   else
      r := '<INPUT TYPE=PASSWORD ';
   r := r + 'NAME=' + Name + ' ';
   r := r + 'VALUE="' + ASCII_To_HTML(Text) + '" ';
   r := r + Format('SIZE=%d', [Size]);
   if MaxLength > 0 then
      r := r + Format('MAXLENGTH=%d', [MaxLength]);
   r := r + '>';
   Result := r;
end;

procedure TWABD_Edit.HTML_To_Object(FormVal: string);
begin
   Text := FormVal;
end;

procedure TWABD_Edit.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FText = '') or (FText=Name)) then Text := Value;
   inherited;
end;

procedure TWABD_Edit.SetSize(NewSize: integer);
begin
   FSize := NewSize;
   FWidth := FSize * PIXELS_PER_CHAR_X + 8;
   FHeight := PIXELS_PER_CHAR_Y + 6;
   Changed;
end;

function TWABD_Edit.Object_To_Control(AOwner: TWinControl): TControl;
var
   ne : TEdit;
begin
   ne := TEdit.Create(AOwner);
   ne.Parent := AOwner;
   ne.Name   := Name;      // Just in case Text is blank, so it doesn't get reset to the Name in SetName
   ne.Text   := Text;
   if IsPassword then ne.PasswordChar := '*';
   Result := ne;
end;


// TWABD_ComboBox

constructor TWABD_ComboBox.Create(AOwner: TComponent);
begin
   inherited;
   FWidth := PIXELS_PER_CHAR_X * 15;
   FHeight := PIXELS_PER_CHAR_Y + 6;
end;

function TWABD_ComboBox.Object_To_HTML: string;
var
   i : integer;
   s : string;
begin
   Result := Format('<SELECT NAME=%s SIZE=%d>'+CR, [Name, 1]);
   for i := 0 to Lines.Count-1 do begin
      if i = SelIndex then s := ' SELECTED' else s := '';
      Result := Result + '<OPTION'+s+'>' + ASCII_To_HTML(Lines[i]) + CR;
   end;
   if Lines.Count = 0 then Result := Result + '<OPTION> ' + CR; // otherwise there will be no ComboBox!
   Result := Result + '</SELECT>'+CR;
end;

procedure TWABD_ComboBox.HTML_To_Object(FormVal: string);
begin
   FSelIndex := Lines.IndexOf(FormVal);
end;

function TWABD_ComboBox.Object_To_Control(AOwner: TWinControl): TControl;
var
   ncb : TComboBox;
begin
   ncb := TComboBox.Create(AOwner);
   ncb.Parent     := AOwner;
   ncb.Name       := Name;
   ncb.Items.Text := Lines.Text;
   ncb.ItemIndex  := SelIndex;
   Result := ncb;
end;


// TWABD_Button

function TWABD_Button.Object_To_HTML: string;
begin
   Result := Format('<INPUT TYPE=SUBMIT NAME=%s VALUE="%s">',
      [WABD_BUTTON_STR, Caption]);
end;

procedure TWABD_Button.HTML_To_Object(FormVal: string);
begin
   raise Exception.Create('WABD Buttons do not have values!');
end;

procedure TWABD_Button.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FCaption='') or (FCaption=Name)) then Caption := Value;
   inherited;
end;

procedure TWABD_Button.SetCaption(NewCaption: string);
begin
   FCaption := NewCaption;
   FWidth    := PIXELS_PER_CHAR_X * Length(FCaption) + 5;
   FHeight   := PIXELS_PER_CHAR_Y + 10;
   Changed;
end;

function TWABD_Button.Object_To_Control(AOwner: TWinControl): TControl;
var
   nb : TButton;
begin
   nb := TButton.Create(AOwner);
   nb.Name    := name;
   nb.Caption := Caption;
   nb.Default := Default;
   Result := nb;
end;


// TWABD_ListBox

constructor TWABD_ListBox.Create(AOwner: TComponent);
begin
   inherited;
   Size := 6;
   FSelIndex := -1;
end;

function TWABD_ListBox.Object_To_HTML: string;
var
   i : integer;
   s : string;
begin
   Result := Format('<SELECT NAME=%s SIZE=%d>'+CR, [Name, Size]);
   for i := 0 to Lines.Count-1 do begin
      if i = SelIndex then s := ' SELECTED' else s := '';
      Result := Result + '<OPTION'+s+'>' + ASCII_To_HTML(Lines[i]) + CR;
   end;
   Result := Result + '</SELECT>'+CR;
end;

procedure TWABD_ListBox.HTML_To_Object(FormVal: string);
begin
   FSelIndex := Lines.IndexOf(FormVal);
end;

procedure TWABD_ListBox.SetSize(NewSize: integer);
begin
   FSize := NewSize;
   FWidth := PIXELS_PER_CHAR_X * 15;
   FHeight := PIXELS_PER_CHAR_Y * FSize + 8;
   Changed;
end;

function TWABD_ListBox.Object_To_Control(AOwner: TWinControl): TControl;
var
   nlb   : TListBox;
begin
   nlb := TListBox.Create(AOwner);
   nlb.Parent     := AOwner;
   nlb.Name       := Name;
   nlb.Items.Text := Lines.Text;
   nlb.ItemIndex  := SelIndex;
   Result := nlb;
end;


// TWABD_RadioButton

function TWABD_RadioButton.Object_To_HTML: string;
var
   ck : string;
begin
   if Checked then ck := 'CHECKED ' else ck := '';
   Result := Format('<INPUT TYPE=RADIO NAME=%s VALUE=%s %s>%s',
      [WABD_RADIO_STR, Name, ck, ASCII_To_HTML(Caption)]);
end;

procedure TWABD_RadioButton.HTML_To_Object(FormVal: string);
begin
   Checked := True;
end;

procedure TWABD_RadioButton.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and ((FCaption='') or (FCaption=Name)) then Caption := Value;
   inherited;
end;

procedure TWABD_RadioButton.SetCaption(NewCaption: string);
begin
   FCaption := NewCaption;
   FWidth := PIXELS_PER_CHAR_X * Length(FCaption) + 15;
   FHeight := PIXELS_PER_CHAR_Y + 4;
   Changed;
end;

function TWABD_RadioButton.Object_To_Control(AOwner: TWinControl): TControl;
var
   nrb : TRadioButton;
begin
   nrb := TRadioButton.Create(AOwner);
   nrb.Parent  := AOwner;
   nrb.Name    := Name;
   nrb.Caption := Caption;
   nrb.Checked := Checked;
   Result := nrb;
end;


// TWABD_CheckBox

function TWABD_CheckBox.Object_To_HTML: string;
var
   ck : string;
begin
   if Checked then ck := 'CHECKED ' else ck := '';
   Result := Format('<INPUT TYPE=CHECKBOX NAME=%s %s', [Name, ck]);
   Result := Result + Format('VALUE=%s>%s', [Name, ASCII_To_HTML(Caption)]);
end;

procedure TWABD_CheckBox.HTML_To_Object(FormVal: string);
begin
   Checked := True;
end;

procedure TWABD_CheckBox.SetName(const Value: TComponentName);
begin
   if (not (csLoading in ComponentState)) and (FCaption = '') then Caption := Value;
   inherited;
end;

procedure TWABD_CheckBox.SetCaption(NewCaption: string);
begin
   FCaption := NewCaption;
   FWidth := PIXELS_PER_CHAR_X * Length(FCaption) + 15;
   FHeight := PIXELS_PER_CHAR_Y + 6;
   Changed;
end;

function TWABD_CheckBox.Object_To_Control(AOwner: TWinControl): TControl;
var
   nck : TCheckBox;
begin
   nck := TCheckBox.Create(AOwner);
   nck.Parent  := AOwner;
   nck.Name    := Name;
   nck.Caption := Caption;
   nck.Checked := Checked;
   Result := nck;
end;


// ************************************************************************
// Utility functions
// ************************************************************************

function FindRep(var s: string; const FindStr, RepStr: string): boolean;
var
   w : integer;
begin
   Result := False;
   if Pos(FindStr, RepStr)<>0 then exit;  // Otherwise it would do infinite recursion

   w := Pos(FindStr, s);
   if (w<>0) then begin
      s := Copy(s, 1, w-1) + RepStr + Copy(s, w + Length(FindStr), Length(s));
      Result := True;
   end;
end;

function FindRepAll(var s: string; const FindStr, RepStr: string): boolean;
begin
   Result := False;
   while FindRep(s, FindStr, RepStr) do
      Result := True;
end;

procedure AmpFixup(var s: string);
var
   tmp : string;
   i   : integer;
begin
   tmp := '';
   for i := 1 to Length(s) do
      if s[i] = '&' then
         tmp := tmp + '&amp'
      else
         tmp := tmp + s[i];
   s := tmp;
end;

function HTML_To_ASCII(const Input: string): string;
begin
   Result := Input;
   FindRepAll(Result, '&amp', '&');
   FindRepAll(Result, '&lt',  '<');
   FindRepAll(Result, '&gt',  '>');
end;

function ASCII_To_HTML(const Input: string): string;
begin
   Result := Input;
   // FindRepAll(Result, '&', '&amp');    // Can't do - inifinite recursion
   AmpFixup(Result);
   FindRepAll(Result, '<', '&lt');
   FindRepAll(Result, '>', '&gt');
end;

function GetLocalBrowserForm: TWABD_Form;
begin
   Result := LocalBrowserForm;
end;


// ************************************************************************
// RegisterWABDComponent - so you can create your own HTML components!
// ************************************************************************

procedure RegisterWABDComponents(WABDClasses: array of TWABD_ObjectClass);
var
   i : integer;
begin
   for i := Low(WABDClasses) to High(WABDClasses) do begin
      RegisterWABDComponent(WABDClasses[i]);
   end;
end;

procedure RegisterWABDComponent(WABDClass: TWABD_ObjectClass);
begin
   if WABDClassList.IndexOf(WABDClass) = -1 then
      WABDClassList.Add(WABDClass);
   RegisterClasses([WABDClass]);
end;

function GetWABDClassListCount: integer;
begin
   Result := WABDClassList.Count;
end;

function GetWABDClassListItem(i: integer): TWABD_ObjectClass;
begin
   if (i < 0) or (i >= GetWABDClassListCount) then
      Result := nil
   else
      Result := WABDClassList.Items[i];
end;


// ************************************************************************
// Registration
// ************************************************************************

procedure Register;
begin
   RegisterComponents('WABD', [
      TWABD_SessionMgr,
      TWABD_Session,
      TWABD_Form]);

   RegisterNoIcon([
      TWABD_FormSection,
      TWABD_Header,
      TWABD_Raw,
      TWABD_Table,
      TWABD_DataTable,
      TWABD_Hidden,
      TWABD_BlankLines,

      TWABD_Anchor,
      TWABD_Image,
      TWABD_LiveImage,
      TWABD_HotSpot,
      TWABD_Label,
      TWABD_Memo,
      TWABD_Button,
      TWABD_Edit,
      TWABD_ComboBox,
      TWABD_RadioButton,
      TWABD_ListBox,
      TWABD_CheckBox]);
end;

initialization
   WABDClassList := TList.Create;
   LiveImageNum  := 1;

   RegisterWABDComponents([
      TWABD_FormSection,
      TWABD_Header,
      TWABD_Raw,
      TWABD_Table,
      TWABD_DataTable,
      TWABD_Hidden,
      TWABD_BlankLines]);

   RegisterClasses([
      TWABD_FormSection,
      TWABD_Header,
      TWABD_Raw,
      TWABD_Table,
      TWABD_DataTable,
      TWABD_Hidden,
      TWABD_BlankLines,

      TWABD_Anchor,
      TWABD_Image,
      TWABD_LiveImage,
      TWABD_HotSpot,
      TWABD_Label,
      TWABD_Memo,
      TWABD_Button,
      TWABD_Edit,
      TWABD_ComboBox,
      TWABD_RadioButton,
      TWABD_ListBox,
      TWABD_CheckBox,

      TWABD_Table_Strings,
      TWABD_HotSpots]);
finalization
   WABDClassList.Free;
end.
